from functools import cached_property
from typing import TYPE_CHECKING, Optional

import redis

from ..conf.config_broker import ConfigBrokerRedis
from ..invocation import DistributedInvocation
from ..util.redis_keys import Key
from .base_broker import BaseBroker

if TYPE_CHECKING:
    from ..app import Pynenc


class RedisQueue:
    def __init__(
        self, app: "Pynenc", client: redis.Redis, name: str, namespace: str = "queue"
    ):
        self.client = client
        self.key = Key(app.app_id, "broker")

    def send_message(self, message: str) -> None:
        self.client.rpush(self.key.default_queue(), message)

    def receive_message(self) -> Optional[str]:
        if msg := self.client.lpop(self.key.default_queue()):
            return msg.decode()
        return None

    def purge(self) -> None:
        self.key.purge(self.client)


class RedisBroker(BaseBroker):
    def __init__(self, app: "Pynenc") -> None:
        super().__init__(app)
        client = redis.Redis(
            host=self.conf.redis_host, port=self.conf.redis_port, db=self.conf.redis_db
        )
        self.queue = RedisQueue(app, client, "default")

    @cached_property
    def conf(self) -> ConfigBrokerRedis:
        return ConfigBrokerRedis(
            config_values=self.app.config_values,
            config_filepath=self.app.config_filepath,
        )

    def route_invocation(self, invocation: "DistributedInvocation") -> None:
        self.queue.send_message(invocation.to_json())

    def retrieve_invocation(self) -> Optional["DistributedInvocation"]:
        if inv := self.queue.receive_message():
            return DistributedInvocation.from_json(self.app, inv)
        return None

    def purge(self) -> None:
        return self.queue.purge()



from __future__ import annotations

import importlib
import json
from dataclasses import dataclass
from functools import cached_property
from typing import TYPE_CHECKING, Any, Generic, Iterable

from .arguments import Arguments
from .call import Call
from .conf.base_config_option import BaseConfigOption
from .conf.single_invocation_pending import SingleInvocation
from .invocation import (
    BaseInvocation,
    BaseInvocationGroup,
    DistributedInvocationGroup,
    SynchronousInvocation,
    SynchronousInvocationGroup,
)
from .types import Func, Params, Result
from .util.log import TaskLoggerAdapter

if TYPE_CHECKING:
    from .app import Pynenc


@dataclass
class TaskOptions:
    """The options common to any implementation of BaseTask"""

    #: If True, only one request will be routed by the broker.
    #: Use this option for tasks that make no sense to execute multiple times in parallel or to avoid generating too much unnecessary tasks in the system.
    single_invocation: SingleInvocation | None = None

    #: If 0 auto parallelization will be disabled.
    #: If > 0, the iterable will be automatically split in chunks of this size and each chunk will be sent to a different worker.
    #: if the task arguments is not an iterable, nothing will happen.
    auto_parallel_batch_size: int = 0

    #: Profiling will take care of storing profiling information for the task (this is a todo, will require further options).
    profiling: str | None = None

    def __post_init__(self) -> None:
        for attr, value in self.__dict__.items():
            if isinstance(value, dict):
                self.__dict__[attr] = BaseConfigOption.from_dict(value)
        self.validate()

    def validate(self) -> None:
        for attr, value in self.__dict__.items():
            if isinstance(value, (str, int, bool, type(None))):
                continue
            if not issubclass(type(value), BaseConfigOption):
                raise TypeError(
                    f"Attribute {attr} must be a basic type or a subclass of BaseConfigOption"
                )

    def to_dict(self) -> dict[str, Any]:
        """Returns a dictionary with the options"""
        options_dict = {**self.__dict__}
        for attr, value in options_dict.items():
            if isinstance(value, BaseConfigOption):
                options_dict[attr] = value.to_dict()
        return options_dict

    def to_json(self) -> str:
        """Returns a string with the serialized options"""
        return json.dumps(self.to_dict())

    @classmethod
    def from_dict(cls, options_dict: dict[str, Any]) -> TaskOptions:
        """Returns a new options from a dictionary"""
        for attr, value in options_dict.items():
            if isinstance(value, dict):
                options_dict[attr] = BaseConfigOption.from_dict(value)
        return cls(**options_dict)

    @classmethod
    def from_json(cls, serialized: str) -> TaskOptions:
        return cls.from_dict(json.loads(serialized))


class Task(Generic[Params, Result]):
    """
    A task in the Pynenc library that represents a function that can be distributed.

    Parameters
    ----------
    app : Pynenc
        A reference to the Pynenc application.
    func : Callable
        The function to be run distributed.
    options : dict
        The options to apply.

    The `BaseTask` can be called normally and will return an instance of `BaseResult`.
    The result will be an `AsyncResult` when running normally but can be `SyncResult`
    when running eagerly in development with the `pynenc` app's `dev_mode_force_sync_tasks`
    option set to `True` (or the 'PYNENC_DEV_MODE_FORCE_SYNC_TASK' environment variable set).
    The option `dev_mode_force_sync_tasks` should only be used in development.

    Although it is possible to create a `BaseTask` instance directly, it is recommended to
    use the decorator provided in the `pynenc` application, i.e., `@app.task(options...)`.
    This is the expected way of instantiating a class and registering it in the app.

    Examples
    --------
    >>> @app.task(options)
    ... def func():
    ...     pass
    ...
    >>> result = func()
    """

    def __init__(self, app: Pynenc, func: Func, options: dict[str, Any]) -> None:
        self.task_id = f"{func.__module__}.{func.__name__}"
        self.app = app
        self.logger = TaskLoggerAdapter(self.app.logger, self.task_id)
        self.func = func
        self.options: TaskOptions = TaskOptions(**options)

    def to_json(self) -> str:
        """Returns a string with the serialized task"""
        return json.dumps({"task_id": self.task_id, "options": self.options.to_json()})

    def __getstate__(self) -> dict:
        # Return state as a dictionary and a secondary value as a tuple
        return {"app": self.app, "task_json": self.to_json()}

    def __setstate__(self, state: dict) -> None:
        # Restore instance attributes
        self.app = state["app"]
        serialized = state["task_json"]
        task_id, func, options = Task._from_json(self.app, serialized)
        # Restore the cached property
        self.task_id = task_id
        self.app = self.app
        self.func = func
        self.options = TaskOptions.from_dict(options)

    @staticmethod
    def _from_json(app: Pynenc, serialized: str) -> tuple[str, Func, dict[str, Any]]:
        """Returns a function and options from a serialized task"""
        task_dict = json.loads(serialized)
        task_id = task_dict["task_id"]
        module_name, function_name = task_id.rsplit(".", 1)
        module = importlib.import_module(module_name)
        function = getattr(module, function_name)
        options_dict = TaskOptions.from_json(task_dict["options"]).to_dict()
        return task_id, function.func, options_dict

    @classmethod
    def from_json(cls, app: Pynenc, serialized: str) -> Task:
        """Returns a new task from a serialized task"""
        _, func, options = cls._from_json(app, serialized)
        return cls(app, func, options)

    @cached_property
    def task_id(self) -> str:
        """The id of the task, which is the module and function name."""
        return f"{self.func.__module__}.{self.func.__name__}"

    def __str__(self) -> str:
        return f"Task(func={self.func.__name__})"

    def __repr__(self) -> str:
        return self.__str__()

    def args(self, *args: Params.args, **kwargs: Params.kwargs) -> Arguments:
        """Returns an Arguments instance from the given args and kwargs"""
        return Arguments.from_call(self.func, *args, **kwargs)

    def __call__(
        self, *args: Params.args, **kwargs: Params.kwargs
    ) -> BaseInvocation[Params, Result]:
        """Handles a call to the task"""
        arguments = Arguments.from_call(self.func, *args, **kwargs)
        return self._call(arguments)

    def _call(self, arguments: Arguments) -> BaseInvocation[Params, Result]:
        """Route the call to the orchestrator if not in dev mode, otherwise run synchronously"""
        if self.app.conf.dev_mode_force_sync_tasks:
            return SynchronousInvocation(
                call=Call(self, arguments),
                result=self.func(**arguments.kwargs),
            )
        return self.app.orchestrator.route_call(Call(self, arguments))

    def parallelize(
        self, param_iter: Iterable[tuple | dict | Arguments]
    ) -> BaseInvocationGroup:
        """
        iterable of calls to the task,
        will accept a tuple positional arguments,
        a dict of keyword arguments,
        or an Arguments instance, eg: task.args(``*args``, ``**kwargs``))
        """
        group_cls: type[BaseInvocationGroup]
        if self.app.conf.dev_mode_force_sync_tasks:
            group_cls = SynchronousInvocationGroup
        else:
            group_cls = DistributedInvocationGroup

        def get_args(params: tuple | dict | Arguments) -> Arguments:
            if isinstance(params, tuple):
                return self.args(*params)
            if isinstance(params, dict):
                return self.args(**params)
            return params

        invocations = []
        for params in param_iter:
            if invocation := self._call(get_args(params)):
                invocations.append(invocation)
        return group_cls(self, invocations)



import builtins
import json
from enum import StrEnum
from typing import Any

from .base_serializer import BaseSerializer


class ReservedKeys(StrEnum):
    """Keys that are reserved for internal use"""

    ERROR = "__pynenc__std_py_exc__"


class DefaultJSONEncoder(json.JSONEncoder):
    def default(self, obj: Any) -> Any:
        if isinstance(obj, Exception):
            return {
                ReservedKeys.ERROR.value: {
                    "type": type(obj).__name__,
                    "args": obj.args,
                    "message": str(obj),
                }
            }
        return super().default(obj)


class JsonSerializer(BaseSerializer):
    @staticmethod
    def serialize(obj: Any) -> str:
        return json.dumps(obj, cls=DefaultJSONEncoder)

    @staticmethod
    def deserialize(obj: str) -> Any:
        data = json.loads(obj)
        if isinstance(data, dict):
            if error_data := data.get(ReservedKeys.ERROR.value):
                error_type = error_data["type"]
                error_args = error_data["args"]
                if hasattr(builtins, error_type):
                    return getattr(builtins, error_type)(*error_args)
        return data



from typing import TYPE_CHECKING

import redis

if TYPE_CHECKING:
    from ..invocation.status import InvocationStatus


def sanitize_for_redis(s: str) -> str:
    replacements = {
        "[": "__OPEN_BRACKET__",
        "]": "__CLOSE_BRACKET__",
        "*": "__ASTERISK__",
    }

    for k, v in replacements.items():
        s = s.replace(k, v)

    return s


class Key:
    def __init__(self, app_id: str, prefix: str) -> None:
        app_id = sanitize_for_redis(app_id)
        prefix = sanitize_for_redis(prefix)
        if not prefix:
            raise ValueError("Prefix cannot be an empty string or None")
        if prefix and not prefix.endswith(":"):
            prefix += ":"
        if app_id:
            prefix = f"{app_id}:{prefix}"
        self.prefix = f"__pynenc__:{prefix}"

    def invocation(self, invocation_id: str) -> str:
        return f"{self.prefix}invocation:{invocation_id}"

    def task(self, task_id: str) -> str:
        return f"{self.prefix}task:{task_id}"

    def args(self, task_id: str, arg: str, val: str) -> str:
        return f"{self.prefix}task:{task_id}:arg:{arg}:val:{val}"

    def status(self, task_id: str, status: "InvocationStatus") -> str:
        return f"{self.prefix}task:{task_id}:status:{status}"

    def pending_timer(self, invocation_id: str) -> str:
        return f"{self.prefix}pending_timer:{invocation_id}"

    def previous_status(self, invocation_id: str) -> str:
        return f"{self.prefix}invocation_previous_status:{invocation_id}"

    def invocation_status(self, invocation_id: str) -> str:
        return f"{self.prefix}invocation_status:{invocation_id}"

    def call(self, call_id: str) -> str:
        return f"{self.prefix}call:{call_id}"

    def call_to_invocation(self, call_id: str) -> str:
        return f"{self.prefix}call_to_invocation:{call_id}"

    def edge(self, call_id: str) -> str:
        return f"{self.prefix}edge:{call_id}"

    def waiting_for(self, invocation_id: str) -> str:
        return f"{self.prefix}waiting_for:{invocation_id}"

    def waited_by(self, invocation_id: str) -> str:
        return f"{self.prefix}waited_by:{invocation_id}"

    def all_waited(self) -> str:
        return f"{self.prefix}all_waited"

    def not_waiting(self) -> str:
        return f"{self.prefix}not_waiting"

    def history(self, invocation_id: str) -> str:
        return f"{self.prefix}history:{invocation_id}"

    def result(self, invocation_id: str) -> str:
        return f"{self.prefix}result:{invocation_id}"

    def exception(self, invocation_id: str) -> str:
        return f"{self.prefix}exception:{invocation_id}"

    def invocation_auto_purge(self) -> str:
        return f"{self.prefix}invocation_auto_purge"

    def default_queue(self) -> str:
        return f"{self.prefix}default_queue"

    def purge(self, client: redis.Redis) -> None:
        """Purge all keys with the given prefix"""
        scan = client.scan_iter(self.prefix + "*")
        for key in scan:
            client.delete(key)



from .base_serializer import BaseSerializer
from .json_serializer import JsonSerializer
from .pickle_serializer import PickleSerializer

__all__ = ["BaseSerializer", "JsonSerializer", "PickleSerializer"]



from functools import cached_property
from logging import Logger
from typing import TYPE_CHECKING, Any, Callable, Optional, Type, overload

from .broker import BaseBroker, MemBroker
from .conf.config_pynenc import ConfigPynenc
from .exceptions import AlreadyInitializedError
from .orchestrator import BaseOrchestrator, MemOrchestrator
from .runner import BaseRunner, DummyRunner
from .serializer import BaseSerializer, JsonSerializer
from .state_backend import BaseStateBackend, MemStateBackend
from .task import Task
from .util.log import create_logger
from .util.subclasses import get_subclass

if TYPE_CHECKING:
    from .types import Func, Params, Result


class Pynenc:
    """
    The main class of the Pynenc library that creates an application object.

    Parameters
    ----------
    task_broker : BaseTaskBroker
        Handles routing of tasks for distributed execution.
    state_backend : BaseStateBackend
        Maintains the state of tasks, runners, and other relevant system states.
    orchestrator : BaseOrchestrator
        Coordinates all components and acts according to the configuration.
    reporting : list of BaseReporting
        Reports to one or more systems.

    Notes
    -----
    All of these base classes are abstract and cannot be used directly. If none is specified,
    they will default to `MemTaskBroker`, `MemStateBackend`, etc. These default classes do not
    actually distribute the code but are helpers for tests or for running an application on your
    localhost. They may help to parallelize to some degree but cannot be used in a production system.

    Examples
    --------
    Default Pynenc application for running in memory in a local environment.

    >>> app = Pynenc()
    """

    _orchestrator_cls: Type[BaseOrchestrator] = MemOrchestrator
    _broker_cls: Type[BaseBroker] = MemBroker
    _state_backend_cls: Type[BaseStateBackend] = MemStateBackend
    _serializer_cls: Type[BaseSerializer] = JsonSerializer
    _runner_cls: Type[BaseRunner] = DummyRunner

    def __init__(
        self,
        app_id: str = "pynenc",
        config_values: Optional[dict[str, Any]] = None,
        config_filepath: Optional[str] = None,
    ) -> None:
        self._app_id = app_id
        self.config_values = config_values
        self.config_filepath = config_filepath
        self.reporting = None
        self._runner_instance: Optional[BaseRunner] = None
        self.logger.info(f"Initialized Pynenc app with id {self.app_id}")

    @property
    def app_id(self) -> str:
        return self._app_id

    def __getstate__(self) -> dict:
        # Return state as a dictionary and a secondary value as a tuple
        return {
            "app_id": self.app_id,
            "orchestrator_cls": self._orchestrator_cls.__name__,
            "broker_cls": self._broker_cls.__name__,
            "state_backend_cls": self._state_backend_cls.__name__,
            "serializer_cls": self._serializer_cls.__name__,
            "runner_cls": self._runner_cls.__name__,
            "config_values": self.config_values,
            "config_filepath": self.config_filepath,
            "reporting": self.reporting,
        }

    def __setstate__(self, state: dict) -> None:
        # Restore instance attributes
        self._app_id = state["app_id"]
        object.__setattr__(self, "_app_id", self._app_id)
        self._orchestrator_cls = get_subclass(
            BaseOrchestrator,  # type: ignore # mypy issue #4717
            state["orchestrator_cls"],
        )
        self._broker_cls = get_subclass(
            BaseBroker,  # type: ignore # mypy issue #4717
            state["broker_cls"],
        )
        self._state_backend_cls = get_subclass(
            BaseStateBackend,  # type: ignore # mypy issue #4717
            state["state_backend_cls"],
        )
        self._serializer_cls = get_subclass(
            BaseSerializer,  # type: ignore # mypy issue #4717
            state["serializer_cls"],
        )
        self._runner_cls = get_subclass(
            BaseRunner,  # type: ignore # mypy issue #4717
            state["runner_cls"],
        )
        self.config_values = state["config_values"]
        self.config_filepath = state["config_filepath"]
        self.reporting = state["reporting"]
        self._runner_instance = None

    def is_initialized(self, property_name: str) -> bool:
        """Returns True if the given cached_property has been initialized"""
        return property_name in self.__dict__

    @cached_property
    def conf(self) -> ConfigPynenc:
        return ConfigPynenc(
            config_values=self.config_values, config_filepath=self.config_filepath
        )

    @cached_property
    def logger(self) -> Logger:
        return create_logger(self)

    @cached_property
    def orchestrator(self) -> BaseOrchestrator:
        return self._orchestrator_cls(self)

    @cached_property
    def broker(self) -> BaseBroker:
        return self._broker_cls(self)

    @cached_property
    def state_backend(self) -> BaseStateBackend:
        return self._state_backend_cls(self)

    @cached_property
    def serializer(self) -> BaseSerializer:
        return self._serializer_cls()

    @property
    def runner(self) -> BaseRunner:
        if self._runner_instance is None:
            self._runner_instance = self._runner_cls(self)
        return self._runner_instance

    @runner.setter
    def runner(self, runner_instance: BaseRunner) -> None:
        self._runner_instance = runner_instance

    def set_orchestrator_cls(self, orchestrator_cls: Type[BaseOrchestrator]) -> None:
        if self.is_initialized(property_name="orchestrator"):
            raise AlreadyInitializedError(
                f"Not possible to set orchestrator instance, already initialized {self._orchestrator_cls}"
            )
        self._orchestrator_cls = orchestrator_cls

    def set_broker_cls(self, broker_cls: Type[BaseBroker]) -> None:
        if self.is_initialized(property_name="broker"):
            raise AlreadyInitializedError(
                f"Not possible to set broker, already initialized {self._broker_cls}"
            )
        self._broker_cls = broker_cls

    def set_state_backend_cls(self, state_backend_cls: Type[BaseStateBackend]) -> None:
        if self.is_initialized(property_name="state_backend"):
            raise AlreadyInitializedError(
                f"Not possible to set state backend, already initialized {self._state_backend_cls}"
            )
        self._state_backend_cls = state_backend_cls

    def set_serializer_cls(self, serializer_cls: Type[BaseSerializer]) -> None:
        if self.is_initialized(property_name="serializer"):
            raise AlreadyInitializedError(
                f"Not possible to set serializer, already initialized {self._serializer_cls}"
            )
        self._serializer_cls = serializer_cls

    def purge(self) -> None:
        """Purge all data from the broker and state backend"""
        self.broker.purge()
        self.orchestrator.purge()
        self.state_backend.purge()

    @overload
    def task(self, func: "Func", **options: Any) -> "Task":
        ...

    @overload
    def task(self, func: None = None, **options: Any) -> Callable[["Func"], "Task"]:
        ...

    def task(
        self, func: Optional["Func"] = None, **options: Any
    ) -> "Task" | Callable[["Func"], "Task"]:
        """
        The task decorator converts the function into an instance of a BaseTask. It accepts any kind of options,
        however these options will be validated with the options class assigned to the class.

        Check the options reference in self.task_cls.options_cls or the Pynenc documentation for a detailed explanation
        of the BaseTask instance you are applying.

        Parameters
        ----------
        func : Callable, optional
            The function to be converted into a BaseTask instance.
        ``**options`` : dict
            The options to be passed to the BaseTask instance.

        Returns
        -------
        Task | Callable[..., Task]
            The BaseTask instance or a callable that returns a BaseTask instance.

        Examples
        --------
        >>> @app.task(option1='value1', option2='value2')
        ... def my_func(x, y):
        ...     return x + y
        ...
        >>> result = my_func(1, 2)
        """

        def init_task(_func: "Func") -> Task["Params", "Result"]:
            if _func.__qualname__ != _func.__name__:
                raise ValueError(
                    "Decorated function must be defined at the module level."
                )
            return Task(self, _func, options)

        if func is None:
            return init_task
        return init_task(func)



"""
Global Pynenc exception and warning classes.
"""
import json
from typing import TYPE_CHECKING, Any, Optional

from .util.subclasses import get_all_subclasses

if TYPE_CHECKING:
    from .call import Call
    from .invocation import BaseInvocation


class PynencError(Exception):
    """Base class for all Pynenc related errors."""

    def _to_json_dict(self) -> dict[str, Any]:
        """Returns a json serializable dictionary"""
        return self.__dict__

    @classmethod
    def _from_json_dict(cls, json_dict: dict[str, Any]) -> "PynencError":
        """Returns a new error from the serialized json compatible dictionary"""
        return cls(**json_dict)

    def to_json(self) -> str:
        """Returns a string with the serialized error"""
        return json.dumps(self._to_json_dict())

    @classmethod
    def from_json(cls, error_name: str, serialized: str) -> "PynencError":
        """Returns the child class from a serialized error"""
        for subcls in get_all_subclasses(cls):
            if subcls.__name__ == error_name:
                return subcls._from_json_dict(json_dict=json.loads(serialized))
        raise ValueError(f"Unknown error type: {error_name}")


class PendingInvocationLockError(PynencError):
    """Error raised when two processes try to set the same invocation as pending concurrently"""

    def __init__(self, invocation_id: str) -> None:
        self.invocation_id = invocation_id


class TaskError(PynencError):
    """Base class for all Task related errors."""

    def __init__(self, task_id: str, message: Optional[str] = None) -> None:
        self.task_id = task_id
        self.message = message

    def __str__(self) -> str:
        if self.message:
            return f"TaskError({self.task_id}): {self.message}"
        else:
            return f"TaskError({self.task_id})"

    def _to_json_dict(self) -> dict[str, Any]:
        """Returns a string with the serialized error"""
        return {"task_id": self.task_id, "message": self.message}

    @classmethod
    def _from_json_dict(cls, json_dict: dict[str, Any]) -> "TaskError":
        """Returns a new error from a serialized error"""
        return cls(json_dict["task_id"], json_dict["message"])


class TaskRoutingError(TaskError):
    """Error raised when a task will not be routed."""


class SingleInvocationWithDifferentArgumentsError(TaskRoutingError):
    """
    Error raised when there is a pending task with different arguments
    than the current task.
    """

    def __init__(
        self,
        task_id: str,
        existing_invocation_id: str,
        new_call_id: str,
        diff: str,
        message: Optional[str] = None,
    ) -> None:
        self.existing_invocation_id = existing_invocation_id
        self.new_call_id = new_call_id
        self.diff = diff
        super().__init__(task_id, message)

    @classmethod
    def from_call_mismatch(
        cls,
        existing_invocation: "BaseInvocation",
        new_call: "Call",
        message: Optional[str] = None,
    ) -> "SingleInvocationWithDifferentArgumentsError":
        return cls(
            existing_invocation.task.task_id,
            existing_invocation.invocation_id,
            new_call.call_id,
            cls.format_difference(existing_invocation.call, new_call),
            message,
        )

    @staticmethod
    def format_difference(existing_call: "Call", new_call: "Call") -> str:
        existing = existing_call.arguments.kwargs
        new = new_call.arguments.kwargs
        common_keys = set(existing.keys()).intersection(new.keys())
        removed_keys = set(existing.keys()).difference(new.keys())
        added_keys = set(new.keys()).difference(existing.keys())

        lines = []

        lines.append("==============================")
        lines.append(f"Differences for {existing_call.task.task_id}:")
        lines.append("==============================")
        lines.append(f"  * Original: {existing}")
        lines.append(f"  * Updated: {new}")
        lines.append("------------------------------")
        lines.append("  * Changes: ")

        for key in common_keys:
            if existing[key] != new[key]:
                lines.append(f"    - {key}: {existing[key]} -> {new[key]}")

        for key in removed_keys:
            lines.append(f"    - {key}: Removed")

        for key in added_keys:
            lines.append(f"    - {key}: Added")

        lines.append("==============================")

        return "\n".join(lines)

    def __str__(self) -> str:
        if self.message:
            return f"SingleInvocationWithDifferentArgumentsError({self.task_id}) {self.message}\n{self.diff}"
        return (
            f"SingleInvocationWithDifferentArgumentsError({self.task_id})\n{self.diff}"
        )

    def _to_json_dict(self) -> dict[str, Any]:
        """Returns a string with the serialized error"""
        return {
            **super()._to_json_dict(),
            "existing_invocation_id": self.existing_invocation_id,
            "new_call_id": self.new_call_id,
            "diff": self.diff,
        }

    @classmethod
    def _from_json_dict(
        cls, json_dict: dict[str, str]
    ) -> "SingleInvocationWithDifferentArgumentsError":
        """Returns a new error from a serialized error"""
        return cls(
            json_dict["task_id"],
            json_dict["existing_invocation_id"],
            json_dict["new_call_id"],
            json_dict["diff"],
            json_dict["message"],
        )


class InvocationError(PynencError):
    """Base class for all Task related errors."""

    def __init__(self, invocation_id: str, message: Optional[str] = None) -> None:
        self.invocation_id = invocation_id
        self.message = message

    def __str__(self) -> str:
        if self.message:
            return f"InvocationError({self.invocation_id}): {self.message}"
        else:
            return f"InvocationError({self.invocation_id})"

    def _to_json_dict(self) -> dict[str, Any]:
        return {"invocation_id": self.invocation_id, "message": self.message}

    @classmethod
    def _from_json_dict(cls, json_dict: dict[str, Any]) -> "InvocationError":
        return cls(json_dict["invocation_id"], json_dict["message"])


class StateBackendError(PynencError):
    """Error raised when a task will not be routed."""


class InvocationNotFoundError(StateBackendError):
    """Error raised when the invocation is not present in the State Backend."""

    def __init__(self, invocation_id: str, message: Optional[str] = None) -> None:
        self.invocation_id = invocation_id
        self.message = message

    def __str__(self) -> str:
        if self.message:
            return f"InvocationNotFoundError({self.invocation_id}): {self.message}"
        else:
            return f"InvocationNotFoundError({self.invocation_id})"

    def _to_json_dict(self) -> dict[str, Any]:
        return {"invocation_id": self.invocation_id, "message": self.message}

    @classmethod
    def _from_json_dict(cls, json_dict: dict[str, Any]) -> "InvocationNotFoundError":
        return cls(json_dict["invocation_id"], json_dict["message"])


class RunnerNotExecutableError(PynencError):
    """Raised when trying to execute a runner that is not meant to be executed."""


class CycleDetectedError(PynencError):
    """Raised when a cycle is detected in the DependencyGraph"""

    def __init__(self, call_ids: list[str], message: str) -> None:
        self.call_ids = call_ids
        self.message = message
        super().__init__(message)

    @classmethod
    def from_cycle(cls, cycle: list["Call"]) -> "CycleDetectedError":
        call_ids = [call.call_id for call in cycle]
        message = f"A cycle was detected: {cls._format_cycle(cycle)}"
        return cls(call_ids, message)

    @staticmethod
    def _format_cycle(cycle: list["Call"]) -> str:
        calls_repr = []
        for call in cycle:
            task = call.task
            func_repr = f"{task.func.__module__}.{task.func.__name__}"
            args_repr = ", ".join(f"{k}:{v}" for k, v in call.arguments.kwargs.items())
            calls_repr.append(f"{func_repr}({args_repr})")

        calls_repr.append(f"back to {calls_repr[0]}")  # Closing the cycle

        formatted_cycle = "\n".join(f"- {call}" for call in calls_repr)

        return f"Cycle detected:\n{formatted_cycle}"

    def _to_json_dict(self) -> dict[str, Any]:
        return {"call_ids": self.call_ids, "message": self.message}

    @classmethod
    def _from_json_dict(cls, json_dict: dict[str, Any]) -> "CycleDetectedError":
        return cls(json_dict["call_ids"], json_dict["message"])


class RunnerError(PynencError):
    """Base class for all Runner related errors."""


class ConfigError(PynencError):
    """Base class for all the config related errors"""


class ConfigMultiInheritanceError(ConfigError):
    """Error related with multiinheritance of config fields"""


class AlreadyInitializedError(PynencError):
    """Error raised when trying to change the class of a component after it was initialized"""



from abc import ABC, abstractmethod
from typing import Any


class BaseSerializer(ABC):
    @staticmethod
    @abstractmethod
    def serialize(obj: Any) -> str:
        ...

    @staticmethod
    @abstractmethod
    def deserialize(obj: str) -> Any:
        ...



from abc import ABC, abstractmethod
from typing import Any, TypeVar

from ..util.subclasses import get_subclass

T = TypeVar("T", bound="BaseConfigOption")


class BaseConfigOption(ABC):
    """Base class for all configuration options."""

    @abstractmethod
    def _to_dict(self) -> dict[str, Any]:
        """Returns a dictionary with the option"""
        pass

    @classmethod
    @abstractmethod
    def _from_dict(cls: type[T], options_dict: dict[str, Any]) -> T:
        """Returns a new option from a dictionary"""
        pass

    def to_dict(self) -> dict[str, Any]:
        """Returns a dictionary with the option"""
        option_dict = self._to_dict()
        option_dict["type"] = self.__class__.__name__
        return option_dict

    @classmethod
    def from_dict(cls: type[T], options_dict: dict[str, Any]) -> T:
        """Returns a new option from a dictionary"""
        _type = options_dict.pop("type")
        subclass = get_subclass(cls, _type)
        return subclass._from_dict(options_dict)



from typing import Any, Optional

from .base_config_option import BaseConfigOption


class SingleInvocation(BaseConfigOption):
    """Only one invocation in status REGISTERED for the specific task will exists in the system

    on_diff_args_raise: bool -> Behaviour when found an existing invocation with different arguments
        - False (Default)       Returns a ReusedInvocation with the different arguments
        - True:                 Raises an exception

    """

    # REGISTERED is the initial status, even before pending
    # when the start gets to run or it's managed by the system it will let new calls be registered

    def __init__(self, on_diff_args_raise: bool = False) -> None:
        self.on_diff_args_raise = on_diff_args_raise

    def get_key_arguments(
        self, serialized_arguments: dict[str, str]
    ) -> Optional[dict[str, str]]:
        del serialized_arguments
        return None

    def _to_dict(self) -> dict[str, Any]:
        """Returns a dictionary with the options"""
        return self.__dict__

    @classmethod
    def _from_dict(cls, options_dict: dict[str, Any]) -> "SingleInvocation":
        """Returns a new options from a dictionary"""
        return cls(**options_dict)


class SingleInvocationPerArguments(SingleInvocation):
    """Only one task request per unique set of arguments can be status:registed in the system"""

    def get_key_arguments(
        self, serialized_arguments: dict[str, str]
    ) -> Optional[dict[str, str]]:
        return serialized_arguments


class SingleInvocationPerKeyArguments(SingleInvocation):
    """Only one task request per unique set of key arguments can be status:registed in the system

    on_diff_args_raise: bool -> Behaviour when found an existing invocation with different arguments
        - False (Default)       Returns a ReusedInvocation with the different arguments
        - True:                 Raises an exception

    """

    def __init__(self, keys: list[str], on_diff_args_raise: bool = False) -> None:
        self.on_diff_args_raise = on_diff_args_raise
        self.keys = keys

    def get_key_arguments(
        self, serialized_arguments: dict[str, str]
    ) -> Optional[dict[str, str]]:
        return {key: serialized_arguments[key] for key in self.keys}



from abc import ABC, abstractmethod
from functools import cached_property
from typing import TYPE_CHECKING, Optional

from ..call import Call
from ..conf.config_broker import ConfigBroker
from ..context import invocation_context
from ..invocation import DistributedInvocation
from ..types import Params, Result

if TYPE_CHECKING:
    from ..app import Pynenc


class BaseBroker(ABC):
    def __init__(self, app: "Pynenc") -> None:
        self.app = app

    @cached_property
    def conf(self) -> ConfigBroker:
        return ConfigBroker(
            config_values=self.app.config_values,
            config_filepath=self.app.config_filepath,
        )

    @abstractmethod
    def route_invocation(self, invocation: DistributedInvocation) -> None:
        ...

    @abstractmethod
    def retrieve_invocation(self) -> Optional[DistributedInvocation]:
        ...

    @abstractmethod
    def purge(self) -> None:
        ...

    # @abstractmethod
    # def _acknowledge_invocation(self, invocation: DistributedInvocation) -> None:
    #     ...

    # @abstractmethod
    # def _requeue_invocation(self, invocation: DistributedInvocation) -> None:
    #     ...

    def route_call(
        self, call: "Call[Params, Result]"
    ) -> DistributedInvocation[Params, Result]:
        """Creates a new invocation and routes it"""
        self.route_invocation(
            invocation := DistributedInvocation(
                call, parent_invocation=invocation_context.get(self.app.app_id)
            )
        )
        self.app.logger.debug(
            f"Routed {call=} on invocation {invocation.invocation_id}"
        )
        return invocation



from .config_base import ConfigBase
from .config_redis import ConfigRedis


class ConfigBroker(ConfigBase):
    """Main config of the boker components"""


class ConfigBrokerRedis(ConfigBroker, ConfigRedis):
    """Specific Configuration for the Redis Broker"""



import json
from dataclasses import dataclass, field
from functools import cached_property
from typing import TYPE_CHECKING, Any, Generic

from .arguments import Arguments
from .types import Params, Result

if TYPE_CHECKING:
    from .app import Pynenc
    from .task import Task


@dataclass(frozen=True)
class Call(Generic[Params, Result]):
    """A specific call of a task

    A call is unique per task and arguments"""

    task: "Task[Params, Result]"
    arguments: "Arguments" = field(default_factory=Arguments)

    @cached_property
    def app(self) -> "Pynenc":
        return self.task.app

    @cached_property
    def call_id(self) -> str:
        """Returns a unique id for each task and arguments"""
        return "#task_id#" + self.task.task_id + "#args_id#" + self.arguments.args_id

    @cached_property
    def serialized_arguments(self) -> dict[str, str]:
        """Returns a dictionary with each of the call arguments serialized into a string"""
        return {
            k: self.app.serializer.serialize(v)
            for k, v in self.arguments.kwargs.items()
        }

    def deserialize_arguments(self, serialized_arguments: dict[str, str]) -> Arguments:
        """Returns an Arguments instance with the deserialized arguments"""
        return Arguments(
            {
                k: self.app.serializer.deserialize(v)
                for k, v in serialized_arguments.items()
            }
        )

    def to_json(self) -> str:
        """Returns a string with the serialized call"""
        return json.dumps(
            {"task": self.task.to_json(), "arguments": self.serialized_arguments}
        )

    def __getstate__(self) -> dict:
        # Return state as a dictionary and a secondary value as a tuple
        return {"task": self.task, "arguments": self.serialized_arguments}

    def __setstate__(self, state: dict) -> None:
        object.__setattr__(self, "task", state["task"])
        arguments = self.deserialize_arguments(state["arguments"])
        object.__setattr__(self, "arguments", arguments)

    @classmethod
    def from_json(cls, app: "Pynenc", serialized: str) -> "Call":
        """Returns a new call from a serialized call"""
        from .task import Task

        call_dict = json.loads(serialized)
        return cls(
            task=Task.from_json(app, call_dict["task"]),
            arguments=Arguments(
                {
                    k: app.serializer.deserialize(v)
                    for k, v in call_dict["arguments"].items()
                }
            ),
        )

    def __str__(self) -> str:
        return f"Call(call_id={self.call_id}, task={self.task}, arguments={self.arguments})"

    def __repr__(self) -> str:
        return self.__str__()

    def __hash__(self) -> int:
        return hash(self.call_id)

    def __eq__(self, other: Any) -> bool:
        # TODO equality based in task and arguments or in invocation_id?
        if not isinstance(other, Call):
            return False
        return self.call_id == other.call_id



from __future__ import annotations

import json
from dataclasses import dataclass
from functools import cached_property
from typing import TYPE_CHECKING, Any, Iterator

from .. import context
from ..arguments import Arguments
from ..call import Call
from ..exceptions import InvocationError
from ..types import Params, Result
from .base_invocation import BaseInvocation, BaseInvocationGroup
from .status import InvocationStatus

if TYPE_CHECKING:
    from ..app import Pynenc


# Create a context variable to store current invocation
@dataclass(frozen=True, eq=False)
class DistributedInvocation(BaseInvocation[Params, Result]):
    """"""

    parent_invocation: DistributedInvocation | None
    _invocation_id: str | None = None

    def __post_init__(self) -> None:
        super().__post_init__()
        self.app.state_backend.upsert_invocation(self)

    @cached_property
    def invocation_id(self) -> str:
        """on deserialization allows to set the invocation_id"""
        return self._invocation_id or super().invocation_id

    @property
    def status(self) -> InvocationStatus:
        """Get the status of the invocation"""
        return self.app.orchestrator.get_invocation_status(self)

    def to_json(self) -> str:
        """Returns a string with the serialized invocation"""
        inv_dict = {"invocation_id": self.invocation_id, "call": self.call.to_json()}
        if self.parent_invocation:
            inv_dict["parent_invocation_id"] = self.parent_invocation.invocation_id
        return json.dumps(inv_dict)

    def __getstate__(self) -> dict:
        # Return state as a dictionary and a secondary value as a tuple
        state = self.__dict__.copy()
        state["invocation_id"] = self.invocation_id
        return state

    def __setstate__(self, state: dict) -> None:
        # Restore instance attributes
        for key, value in state.items():
            object.__setattr__(self, key, value)

    @classmethod
    def from_json(cls, app: Pynenc, serialized: str) -> DistributedInvocation:
        """Returns a new invocation from a serialized invocation"""
        inv_dict = json.loads(serialized)
        call = Call.from_json(app, inv_dict["call"])
        parent_invocation = None
        if "parent_invocation_id" in inv_dict:
            parent_invocation = app.state_backend.get_invocation(
                inv_dict["parent_invocation_id"]
            )
        return cls(call, parent_invocation, inv_dict["invocation_id"])

    def run(self, runner_args: dict[str, Any] | None = None) -> None:
        context.runner_args = runner_args
        # Set current invocation
        previous_invocation_context = context.invocation_context.get(self.app.app_id)
        try:
            context.invocation_context[self.app.app_id] = self
            self.app.orchestrator.set_invocation_run(self.parent_invocation, self)
            result = self.task.func(**self.arguments.kwargs)
            self.app.orchestrator.set_invocation_result(self, result)
        except Exception as ex:
            self.app.orchestrator.set_invocation_exception(self, ex)
            raise ex
        finally:
            context.invocation_context[self.app.app_id] = previous_invocation_context

    @property
    def result(self) -> Result:
        if not self.status.is_final():
            self.app.orchestrator.waiting_for_results(self.parent_invocation, [self])

        while not self.status.is_final():
            self.app.runner.waiting_for_results(
                self.parent_invocation, [self], context.runner_args
            )
        return self.get_final_result()

    def get_final_result(self) -> Result:
        if not self.status.is_final():
            raise InvocationError(self.invocation_id, "Invocation is not final")
        if self.status == InvocationStatus.FAILED:
            raise self.app.state_backend.get_exception(self)
        return self.app.state_backend.get_result(self)


class DistributedInvocationGroup(
    BaseInvocationGroup[Params, Result, DistributedInvocation]
):
    @property
    def results(self) -> Iterator[Result]:
        waiting_invocations = self.invocations.copy()
        if not waiting_invocations:
            return
        parent_invocation = waiting_invocations[0].parent_invocation
        notified_orchestrator = False
        while waiting_invocations:
            for invocation in waiting_invocations:
                if invocation.status.is_final():
                    waiting_invocations.remove(invocation)
                    yield invocation.result
            if not notified_orchestrator:
                self.app.orchestrator.waiting_for_results(
                    parent_invocation, waiting_invocations
                )
            self.app.runner.waiting_for_results(parent_invocation, waiting_invocations)


@dataclass(frozen=True)
class ReusedInvocation(DistributedInvocation):
    """This is an invocation referencing an older one"""

    # Due to single invocation functionality
    # keeps existing invocation + new argument if any change
    diff_arg: Arguments | None = None

    @classmethod
    def from_existing(
        cls, invocation: DistributedInvocation, diff_arg: Arguments | None = None
    ) -> ReusedInvocation:
        # Create a new instance with the same fields as the existing invocation, but with the added field
        new_invc = cls(
            call=Call(invocation.task, invocation.arguments),
            parent_invocation=invocation.parent_invocation,
            # we reuse invocation_id from original invocation
            _invocation_id=invocation.invocation_id,
            diff_arg=diff_arg,
        )
        return new_invc



from .base_runner import BaseRunner, DummyRunner
from .context import ApplicationContext
from .mem_runner import MemRunner
from .process_runner import ProcessRunner

__all__ = [
    "BaseRunner",
    "DummyRunner",
    "MemRunner",
    "ProcessRunner",
    "ApplicationContext",
]



from typing import TypeVar

T = TypeVar("T", bound=object)  # Declare type variable


def get_all_subclasses(cls: type[T]) -> list[type[T]]:
    all_subclasses = []
    # remove all related type:ignores when mypy fix the issue
    # https://github.com/python/mypy/issues/4717
    # type: ignore # mypy issue #4717
    for subclass in cls.__subclasses__():
        all_subclasses.append(subclass)
        all_subclasses.extend(get_all_subclasses(subclass))

    return all_subclasses


def get_subclass(root_class: type[T], child_class_name: str) -> type[T]:
    """Returns the subclass with the given name (any level deep)"""
    for subclass in get_all_subclasses(root_class):
        if subclass.__name__ == child_class_name:
            return subclass
    raise ValueError(f"Unknown subclass: {child_class_name} of {root_class.__name__}")



from enum import StrEnum, auto


class InvocationStatus(StrEnum):
    """
    An enumeration representing the status of a task invocation.

    The PENDING status will expire after the time specified in
    :attr:`~pynenc.conf.config_pynenc.ConfigPynenc.max_pending_seconds`.

    Notes
    -----
    FAILED, RETRY, and SUCCESS are the same category and can be considered as subtypes of a hypothetical TERMINATED status.
    """

    #: The task call has been routed and is registered
    REGISTERED = auto()

    #: The task call was picked by a runner but is not yet executed.
    #: The status pending will expire after Config.max_pending_seconds
    PENDING = auto()

    #: The task call is currently running.
    RUNNING = auto()

    #: The task call finished without errors.
    SUCCESS = auto()

    #: The task call execution is paused.
    PAUSED = auto()

    #: A task has been registered to run at a specific time. This is a subtype of REGISTERED.
    SCHEDULED = auto()

    #: The task call finished with exceptions.
    FAILED = auto()

    #: The task call finished with a retriable exception.
    RETRY = auto()

    def is_available_for_run(self) -> bool:
        """
        Check if the task is in a state where it is available to be picked up and run by any broker.
        This means the task is not currently being executed, paused, or about to be executed by any other broker.

        Returns
        -------
        bool
            True if the status is runnable by any broker, False otherwise.
        """
        return self in {
            InvocationStatus.REGISTERED,
            InvocationStatus.RETRY,
        }

    def is_final(self) -> bool:
        """
        Checks if the status is a final status.

        Returns
        -------
        bool
            True if the status is final, False otherwise.
        """
        return self in {InvocationStatus.SUCCESS, InvocationStatus.FAILED}



from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING, Iterator

from ..exceptions import PynencError
from ..types import Params, Result
from .base_invocation import BaseInvocation, BaseInvocationGroup

if TYPE_CHECKING:
    from ..app import Pynenc
    from ..call import Call


class SynchronousInvocation(BaseInvocation[Params, Result]):
    def __init__(self, call: Call[Params, Result], result: Result) -> None:
        super().__init__(call)
        self._result = result

    @cached_property
    def result(self) -> Result:
        return self._result

    def to_json(self) -> str:
        raise PynencError("SynchronousInvocation cannot be serialized")

    @classmethod
    def from_json(cls, app: Pynenc, serialized: str) -> SynchronousInvocation:
        del app, serialized
        raise PynencError("SynchronousInvocation cannot be deserialized")


class SynchronousInvocationGroup(
    BaseInvocationGroup[Params, Result, SynchronousInvocation]
):
    @property
    def results(self) -> Iterator[Result]:
        for invocation in self.invocations:
            yield invocation.result



from .base_invocation import BaseInvocation, BaseInvocationGroup
from .dist_invocation import (
    DistributedInvocation,
    DistributedInvocationGroup,
    ReusedInvocation,
)
from .status import InvocationStatus
from .sync_invocation import SynchronousInvocation, SynchronousInvocationGroup

__all__ = [
    "BaseInvocation",
    "BaseInvocationGroup",
    "SynchronousInvocation",
    "SynchronousInvocationGroup",
    "DistributedInvocation",
    "ReusedInvocation",
    "DistributedInvocationGroup",
    "InvocationStatus",
]



import codecs
import pickle
from typing import Any

from .base_serializer import BaseSerializer


class PickleSerializer(BaseSerializer):
    @staticmethod
    def serialize(obj: Any) -> str:
        return codecs.encode(pickle.dumps(obj), "base64").decode()

    @staticmethod
    def deserialize(serialized_obj: str) -> Any:
        return pickle.loads(codecs.decode(serialized_obj.encode(), "base64"))



from collections import deque
from typing import TYPE_CHECKING, Optional

from .base_broker import BaseBroker

if TYPE_CHECKING:
    from ..app import Pynenc
    from ..invocation import DistributedInvocation


class MemBroker(BaseBroker):
    def __init__(self, app: "Pynenc") -> None:
        self._queue: deque = deque()
        super().__init__(app)

    def route_invocation(self, invocation: "DistributedInvocation") -> None:
        self._queue.append(invocation)

    def retrieve_invocation(self) -> Optional["DistributedInvocation"]:
        return self._queue.pop() if self._queue else None

    def purge(self) -> None:
        return self._queue.clear()



import logging
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from ..app import Pynenc


def create_logger(app: "Pynenc") -> logging.Logger:
    """Creates a logger for the given app"""
    logger = logging.getLogger(f"pynenc.{app.app_id}")
    if level_name := app.conf.logging_level:
        numeric_level = getattr(logging, app.conf.logging_level.upper(), None)
        if not isinstance(numeric_level, int):
            raise ValueError(f"Invalid log level: {level_name}")
        logger.setLevel(numeric_level)
    return logger


class TaskLoggerAdapter(logging.LoggerAdapter):
    def __init__(
        self, logger: logging.Logger, task_id: str, invocation_id: str | None = None
    ):
        super().__init__(logger, {})
        self.set_context(task_id, invocation_id)

    def set_context(self, task_id: str, invocation_id: str | None) -> None:
        self.task_id = task_id
        self.invocation_id = invocation_id

    def process(self, msg: Any, kwargs: Any) -> Any:
        # prefix = f"[{self.task_id}({self.invocation_id})]"
        if self.invocation_id:
            prefix = f"[{self.task_id}: {self.invocation_id}]"
        else:
            prefix = f"[{self.task_id}]"
        return f"{prefix} {msg}", kwargs


class RunnerLogAdapter(logging.LoggerAdapter):
    def __init__(self, logger: logging.Logger, runner_id: str):
        super().__init__(logger, {})
        self.runner_id = runner_id

    def process(self, msg: Any, kwargs: Any) -> Any:
        prefix = f"[runner: {self.runner_id}]"
        return f"{prefix} {msg}", kwargs



import json
import tomllib
from typing import Any

import yaml


def load_config_from_toml(file_path: str) -> dict:
    with open(file_path, "rb") as toml_file:
        if config_data := tomllib.load(toml_file):
            if "pyproject.toml" in file_path:
                return config_data.get("tool", {}).get("pynenc", {})
        return config_data or {}


def load_file(filepath: str) -> dict[str, Any]:
    with open(filepath) as _file:
        if filepath.lower().endswith(".yaml") or filepath.lower().endswith(".yml"):
            # TODO fix yaml
            return yaml.load(_file, Loader=yaml.SafeLoader)
        if filepath.lower().endswith(".json"):
            return json.load(_file)
        if filepath.lower().endswith(".toml"):
            return load_config_from_toml(filepath)
    raise ValueError(f"Unexpected file extension {filepath=}")



from .base_state_backend import BaseStateBackend
from .mem_state_backend import MemStateBackend
from .redis_state_backend import RedisStateBackend

__all__ = ["BaseStateBackend", "MemStateBackend", "RedisStateBackend"]



import typing

if typing.TYPE_CHECKING:
    from typing_extensions import ParamSpec

    Params = ParamSpec("Params")
else:
    Params = typing.TypeVar("Params")

Result = typing.TypeVar("Result")
Func = typing.Callable[Params, Result]
Args = typing.Dict[str, typing.Any]



from .config_base import ConfigBase, ConfigField
from .config_redis import ConfigRedis


class ConfigOrchestrator(ConfigBase):
    """Main config of the orchestrator components"""

    cycle_control = ConfigField(True)
    blocking_control = ConfigField(True)
    auto_final_invocation_purge_hours = ConfigField(24.0)


class ConfigOrchestratorRedis(ConfigOrchestrator, ConfigRedis):
    """Specific Configuration for the Redis Orchestrator"""



import json
import threading
from abc import ABC, abstractmethod
from collections import defaultdict
from dataclasses import dataclass, field
from datetime import datetime
from functools import cached_property
from typing import TYPE_CHECKING, Any, Optional

from ..conf.config_state_backend import ConfigStateBackend
from ..exceptions import InvocationNotFoundError
from ..invocation import InvocationStatus

if TYPE_CHECKING:
    from ..app import Pynenc
    from ..invocation import DistributedInvocation
    from ..types import Params, Result


@dataclass
class InvocationHistory:
    invocation_id: str
    _timestamp: datetime = field(init=False, default_factory=lambda: datetime.utcnow())
    status: Optional["InvocationStatus"] = None
    execution_context: Optional[Any] = None  # Todo on Runners

    @property
    def timestamp(self) -> datetime:
        return self._timestamp

    def to_json(self) -> str:
        return json.dumps(
            {
                "invocation_id": self.invocation_id,
                "_timestamp": self._timestamp.isoformat(),
                "status": self.status.value if self.status else None,
                "execution_context": self.execution_context,  # TODO
            }
        )

    @classmethod
    def from_json(cls, json_str: str) -> "InvocationHistory":
        hist_dict = json.loads(json_str)
        history = cls(hist_dict["invocation_id"])
        history._timestamp = datetime.fromisoformat(hist_dict["_timestamp"])
        history.status = InvocationStatus(hist_dict["status"])
        history.execution_context = hist_dict["execution_context"]
        return history


class BaseStateBackend(ABC):
    def __init__(self, app: "Pynenc") -> None:
        self.app = app
        self.invocation_threads: dict[str, list[threading.Thread]] = defaultdict(list)

    @cached_property
    def conf(self) -> ConfigStateBackend:
        return ConfigStateBackend(
            config_values=self.app.config_values,
            config_filepath=self.app.config_filepath,
        )

    def wait_for_all_async_operations(self) -> None:
        """Blocks until all asynchronous status operations are finished."""
        for invocation_id in self.invocation_threads:
            self.wait_for_invocation_async_operations(invocation_id)

    def wait_for_invocation_async_operations(self, invocation_id: str) -> None:
        """Blocks until all asynchronous operations for a specific invocation are finished."""
        for thread in self.invocation_threads[invocation_id]:
            thread.join()

    @abstractmethod
    def purge(self) -> None:
        ...

    @abstractmethod
    def _upsert_invocation(
        self, invocation: "DistributedInvocation[Params, Result]"
    ) -> None:
        ...

    @abstractmethod
    def _get_invocation(self, invocation_id: str) -> Optional["DistributedInvocation"]:
        ...

    @abstractmethod
    def _add_history(
        self,
        invocation: "DistributedInvocation[Params, Result]",
        invocation_history: InvocationHistory,
    ) -> None:
        ...

    @abstractmethod
    def _get_history(
        self, invocation: "DistributedInvocation[Params, Result]"
    ) -> list[InvocationHistory]:
        ...

    @abstractmethod
    def _get_result(
        self, invocation: "DistributedInvocation[Params, Result]"
    ) -> "Result":
        ...

    @abstractmethod
    def _set_result(
        self, invocation: "DistributedInvocation[Params, Result]", result: "Result"
    ) -> None:
        ...

    @abstractmethod
    def _get_exception(
        self, invocation: "DistributedInvocation[Params, Result]"
    ) -> "Exception":
        ...

    @abstractmethod
    def _set_exception(
        self,
        invocation: "DistributedInvocation[Params, Result]",
        exception: "Exception",
    ) -> None:
        ...

    def upsert_invocation(self, invocation: "DistributedInvocation") -> None:
        thread = threading.Thread(target=self._upsert_invocation, args=(invocation,))
        thread.start()
        self.invocation_threads[invocation.invocation_id].append(thread)

    def get_invocation(self, invocation_id: str) -> "DistributedInvocation":
        if invocation := self._get_invocation(invocation_id):
            return invocation
        raise InvocationNotFoundError(invocation_id, "The invocation wasn't stored")

    def add_history(
        self,
        invocation: "DistributedInvocation[Params, Result]",
        status: Optional["InvocationStatus"] = None,
        execution_context: Optional["Any"] = None,
    ) -> None:
        invocation_history = InvocationHistory(
            invocation.invocation_id, status, execution_context
        )
        thread = threading.Thread(
            target=self._add_history, args=(invocation, invocation_history)
        )
        self.invocation_threads[invocation.invocation_id].append(thread)
        thread.start()

    def get_history(
        self, invocation: "DistributedInvocation[Params, Result]"
    ) -> list[InvocationHistory]:
        # todo fork open threads
        return self._get_history(invocation)

    def set_result(self, invocation: "DistributedInvocation", result: "Result") -> None:
        self._set_result(invocation, result)

    def get_result(
        self, invocation: "DistributedInvocation[Params, Result]"
    ) -> "Result":
        # insert result is block, no need for thread control
        return self._get_result(invocation)

    def set_exception(
        self, invocation: "DistributedInvocation", exception: "Exception"
    ) -> None:
        self._set_exception(invocation, exception)

    def get_exception(
        self, invocation: "DistributedInvocation[Params, Result]"
    ) -> Exception:
        # todo
        return self._get_exception(invocation)



import json
from functools import cached_property
from typing import TYPE_CHECKING

import redis

from .. import exceptions
from ..conf.config_state_backend import ConfigStateBackendRedis
from ..invocation import DistributedInvocation
from ..util.redis_keys import Key
from .base_state_backend import BaseStateBackend, InvocationHistory

if TYPE_CHECKING:
    from ..app import Pynenc
    from ..types import Params, Result


class RedisStateBackend(BaseStateBackend):
    def __init__(self, app: "Pynenc") -> None:
        super().__init__(app)
        self.client = redis.Redis(
            host=self.conf.redis_host, port=self.conf.redis_port, db=self.conf.redis_db
        )
        self.key = Key(app.app_id, "state_backend")
        # self._cache: dict[str, "DistributedInvocation"] = {}
        # self._history: dict[str, list] = defaultdict(list)
        # self._results: dict[str, Any] = {}
        # self._exceptions: dict[str, Exception] = {}

    @cached_property
    def conf(self) -> ConfigStateBackendRedis:
        return ConfigStateBackendRedis(
            config_values=self.app.config_values,
            config_filepath=self.app.config_filepath,
        )

    def purge(self) -> None:
        self.key.purge(self.client)

    def _upsert_invocation(self, invocation: "DistributedInvocation") -> None:
        self.client.set(
            self.key.invocation(invocation.invocation_id), invocation.to_json()
        )

    def _get_invocation(
        self, invocation_id: str
    ) -> "DistributedInvocation[Params, Result]":
        if inv := self.client.get(self.key.invocation(invocation_id)):
            return DistributedInvocation.from_json(self.app, inv.decode())
        raise KeyError(f"Invocation {invocation_id} not found")

    def _add_history(
        self,
        invocation: "DistributedInvocation",
        invocation_history: "InvocationHistory",
    ) -> None:
        self.client.rpush(
            self.key.history(invocation.invocation_id),
            invocation_history.to_json(),
        )

    def _get_history(
        self, invocation: "DistributedInvocation[Params, Result]"
    ) -> list[InvocationHistory]:
        return [
            InvocationHistory.from_json(h.decode())
            for h in self.client.lrange(
                self.key.history(invocation.invocation_id), 0, -1
            )
        ]

    def _set_result(
        self, invocation: "DistributedInvocation[Params, Result]", result: "Result"
    ) -> None:
        self.client.set(
            self.key.result(invocation.invocation_id),
            self.app.serializer.serialize(result),
        )

    def _get_result(
        self, invocation: "DistributedInvocation[Params, Result]"
    ) -> "Result":
        # return self._results[invocation.invocation_id]
        if res := self.client.get(self.key.result(invocation.invocation_id)):
            return self.app.serializer.deserialize(res.decode())
        raise KeyError(f"Result for invocation {invocation.invocation_id} not found")

    def _set_exception(
        self,
        invocation: "DistributedInvocation[Params, Result]",
        exception: "Exception",
    ) -> None:
        serialized_exception: dict[str, str | bool] = {
            "error_name": exception.__class__.__name__
        }
        if isinstance(exception, exceptions.PynencError):
            serialized_exception["pynenc_error"] = True
            serialized_exception["error_data"] = exception.to_json()
        else:
            serialized_exception["pynenc_error"] = False
            serialized_exception["error_data"] = self.app.serializer.serialize(
                exception
            )
        self.client.set(
            self.key.exception(invocation.invocation_id),
            json.dumps(serialized_exception),
        )

    def _get_exception(
        self, invocation: "DistributedInvocation[Params, Result]"
    ) -> Exception:
        if exc := self.client.get(self.key.exception(invocation.invocation_id)):
            serialized_exception = json.loads(exc.decode())
            if serialized_exception["pynenc_error"]:
                return exceptions.PynencError.from_json(
                    serialized_exception["error_name"],
                    serialized_exception["error_data"],
                )
            return self.app.serializer.deserialize(serialized_exception["error_data"])
        raise KeyError(f"Exception for invocation {invocation.invocation_id} not found")



from abc import ABC, abstractmethod
from functools import cached_property
from typing import TYPE_CHECKING, Any, Iterator, Optional

from ..conf.config_orchestrator import ConfigOrchestrator
from ..context import invocation_context
from ..exceptions import (
    PendingInvocationLockError,
    SingleInvocationWithDifferentArgumentsError,
)
from ..invocation import DistributedInvocation, InvocationStatus, ReusedInvocation

if TYPE_CHECKING:
    from ..app import Pynenc
    from ..call import Call
    from ..task import Task
    from ..types import Params, Result


class BaseCycleControl(ABC):
    """Sub component of the orchestrator to implement cycle control functionalities"""

    @abstractmethod
    def add_call_and_check_cycles(
        self,
        caller_invocation: "DistributedInvocation[Params, Result]",
        callee_invocation: "DistributedInvocation[Params, Result]",
    ) -> None:
        """Adds a new call between invocations and raise an exception to prevent the formation of a call cycle"""
        # TODO async store of call dependencies in state backend

    @abstractmethod
    def clean_up_invocation_cycles(self, invocation: "DistributedInvocation") -> None:
        """Called when an invocation is finished and therefore cannot be part of a cycle anymore"""


class BaseBlockingControl(ABC):
    """Sub component of the orchestrator to implement blocking control functionalities"""

    @abstractmethod
    def release_waiters(self, waited: "DistributedInvocation") -> None:
        """Called when an invocation is finished and therefore cannot block other invocations anymore"""

    @abstractmethod
    def waiting_for_results(
        self,
        caller_invocation: "DistributedInvocation[Params, Result]",
        result_invocations: list["DistributedInvocation[Params, Result]"],
    ) -> None:
        """Called when an Optional[invocation] is waiting in the result result of another invocation."""

    @abstractmethod
    def get_blocking_invocations(
        self, max_num_invocations: int
    ) -> Iterator["DistributedInvocation"]:
        """Returns an iterator of invocations that are blocking other invocations
        but are not getting blocked by any invocation.
        order by age, the oldest invocation first.
        """


class BaseOrchestrator(ABC):
    def __init__(self, app: "Pynenc") -> None:
        self.app = app

    @cached_property
    def conf(self) -> ConfigOrchestrator:
        return ConfigOrchestrator(
            config_values=self.app.config_values,
            config_filepath=self.app.config_filepath,
        )

    @abstractmethod
    def get_existing_invocations(
        self,
        task: "Task[Params, Result]",
        key_serialized_arguments: Optional[dict[str, str]] = None,
        status: Optional["InvocationStatus"] = None,
    ) -> Iterator["DistributedInvocation"]:
        ...

    @abstractmethod
    def _set_invocation_status(
        self,
        invocation: "DistributedInvocation[Params, Result]",
        status: "InvocationStatus",
    ) -> None:
        ...

    @abstractmethod
    def _set_invocation_pending_status(
        self,
        invocation: "DistributedInvocation[Params, Result]",
    ) -> None:
        """Pending can only be set by the orchestrator"""
        ...

    @abstractmethod
    def set_up_invocation_auto_purge(
        self,
        invocation: "DistributedInvocation[Params, Result]",
    ) -> None:
        """set up the invocation to be auto purgue after app.conf.orchestrator_auto_final_invocation_purge_hours"""

    @abstractmethod
    def auto_purge(self) -> None:
        """Purge all invocations in final state that are older than app.conf.orchestrator_auto_final_invocation_purge_hours"""

    @abstractmethod
    def get_invocation_status(
        self, invocation: "DistributedInvocation[Params, Result]"
    ) -> "InvocationStatus":
        ...
        # TODO if invocation does not exists try in state backend (and cache it up)
        #      before raising an exception!!!

    @abstractmethod
    def purge(self) -> None:
        ...

    #############################################
    # cycle sub functionalities
    @property
    @abstractmethod
    def cycle_control(self) -> BaseCycleControl:
        ...

    def add_call_and_check_cycles(
        self,
        caller_invocation: "DistributedInvocation[Params, Result]",
        callee_invocation: "DistributedInvocation[Params, Result]",
    ) -> None:
        """Adds a new call between invocations and raise an exception to prevent the formation of a call cycle"""
        if self.conf.cycle_control:
            self.cycle_control.add_call_and_check_cycles(
                caller_invocation, callee_invocation
            )
        # TODO async store of call dependencies in state backend

    def clean_up_invocation_cycles(self, invocation: "DistributedInvocation") -> None:
        """Called when an invocation is finished and therefore cannot be part of a cycle anymore"""
        if self.conf.cycle_control:
            self.cycle_control.clean_up_invocation_cycles(invocation)

    #############################################
    # blocking sub functionalities
    @property
    @abstractmethod
    def blocking_control(self) -> BaseBlockingControl:
        ...

    def release_waiters(self, waited: "DistributedInvocation") -> None:
        """Called when an invocation is finished and therefore cannot block other invocations anymore"""
        if self.app.orchestrator.conf.blocking_control:
            self.blocking_control.release_waiters(waited)

    def waiting_for_results(
        self,
        caller_invocation: Optional["DistributedInvocation[Params, Result]"],
        result_invocations: list["DistributedInvocation[Params, Result]"],
    ) -> None:
        """Called when an Optional[invocation] is waiting in the result result of another invocation."""
        if self.app.orchestrator.conf.blocking_control and caller_invocation:
            self.blocking_control.waiting_for_results(
                caller_invocation, result_invocations
            )

    def get_blocking_invocations(
        self, max_num_invocations: int
    ) -> Iterator["DistributedInvocation"]:
        """Returns an iterator of invocations that are blocking other invocations
        but are not getting blocked by any invocation.
        order by age, the oldest invocation first.
        """
        if self.app.orchestrator.conf.blocking_control:
            yield from self.blocking_control.get_blocking_invocations(
                max_num_invocations
            )

    #############################################

    def set_invocation_status(
        self,
        invocation: "DistributedInvocation[Params, Result]",
        status: "InvocationStatus",
    ) -> None:
        if status == InvocationStatus.PENDING:
            self._set_invocation_pending_status(invocation)
        else:
            # TODO async store of status in state backend
            if status.is_final():
                self.release_waiters(invocation)
                self.clean_up_invocation_cycles(invocation)
                self.set_up_invocation_auto_purge(invocation)
            # TODO! on previous fail, this should still change status to running
            self._set_invocation_status(invocation, status)

    def set_invocations_status(
        self,
        invocations: list["DistributedInvocation[Params, Result]"],
        status: "InvocationStatus",
    ) -> None:
        for invocation in invocations:
            self.set_invocation_status(invocation, status)

    def set_invocation_run(
        self,
        caller: Optional["DistributedInvocation[Params, Result]"],
        callee: "DistributedInvocation[Params, Result]",
    ) -> None:
        """Called when an invocation is started"""
        if caller:
            self.add_call_and_check_cycles(caller, callee)
        # TODO! on previous fail, this should still change status
        self.set_invocation_status(callee, InvocationStatus.RUNNING)

    def set_invocation_result(
        self, invocation: "DistributedInvocation", result: Any
    ) -> None:
        """Called when an invocation is finished successfully"""
        self.app.state_backend.set_result(invocation, result)
        # TODO! on previous fail, this should still change status
        self.app.orchestrator.set_invocation_status(
            invocation, InvocationStatus.SUCCESS
        )

    def set_invocation_exception(
        self, invocation: "DistributedInvocation", exception: Exception
    ) -> None:
        """Called when an invocation is finished with an exception"""
        self.app.state_backend.set_exception(invocation, exception)
        # TODO! on previous fail, this should still change status
        self.app.orchestrator.set_invocation_status(invocation, InvocationStatus.FAILED)

    def get_invocations_to_run(
        self, max_num_invocations: int
    ) -> Iterator["DistributedInvocation"]:
        """Returns an iterator of max_num_invocations that are ready for running"""
        # first get the blocking invocations
        blocking_invocation_ids: set[str] = set()
        for blocking_invocation in self.get_blocking_invocations(max_num_invocations):
            blocking_invocation_ids.add(blocking_invocation.invocation_id)
            try:
                self._set_invocation_pending_status(blocking_invocation)
                yield blocking_invocation
            except PendingInvocationLockError:
                continue
        missing_invocations = max_num_invocations - len(blocking_invocation_ids)
        # then get the rest from the broker
        while missing_invocations > 0:
            if invocation := self.app.broker.retrieve_invocation():
                if invocation.invocation_id not in blocking_invocation_ids:
                    if self.get_invocation_status(invocation).is_available_for_run():
                        missing_invocations -= 1
                        try:
                            self._set_invocation_pending_status(invocation)
                            yield invocation
                        except PendingInvocationLockError:
                            continue
            else:
                break

    def _route_new_call_invocation(
        self, call: "Call[Params, Result]"
    ) -> "DistributedInvocation[Params, Result]":
        new_invocation = DistributedInvocation(
            call, parent_invocation=invocation_context.get(self.app.app_id)
        )
        self.set_invocation_status(new_invocation, InvocationStatus.REGISTERED)
        self.app.broker.route_invocation(new_invocation)
        return new_invocation

    def route_call(self, call: "Call") -> "DistributedInvocation[Params, Result]":
        if not call.task.options.single_invocation:
            return self._route_new_call_invocation(call)
        # Handleling single invocation routings
        invocation = next(
            self.get_existing_invocations(
                task=call.task,
                key_serialized_arguments=call.task.options.single_invocation.get_key_arguments(
                    call.serialized_arguments
                ),
                status=InvocationStatus.REGISTERED,
            ),
            None,
        )
        if not invocation:
            return self._route_new_call_invocation(call)
        if invocation.serialized_arguments == call.serialized_arguments:
            return ReusedInvocation.from_existing(invocation)
        if call.task.options.single_invocation.on_diff_args_raise:
            raise SingleInvocationWithDifferentArgumentsError.from_call_mismatch(
                existing_invocation=invocation, new_call=call
            )
        return ReusedInvocation.from_existing(invocation, call.arguments)



from collections import defaultdict
from typing import TYPE_CHECKING, Any

from pynenc.state_backend.base_state_backend import InvocationHistory

from .base_state_backend import BaseStateBackend

if TYPE_CHECKING:
    from ..app import Pynenc
    from ..invocation import DistributedInvocation
    from ..types import Params, Result


class MemStateBackend(BaseStateBackend):
    def __init__(self, app: "Pynenc") -> None:
        self._cache: dict[str, "DistributedInvocation"] = {}
        self._history: dict[str, list] = defaultdict(list)
        self._results: dict[str, Any] = {}
        self._exceptions: dict[str, Exception] = {}
        super().__init__(app)

    def purge(self) -> None:
        self._cache.clear()
        self._history.clear()
        self._results.clear()
        self._exceptions.clear()

    def _upsert_invocation(self, invocation: "DistributedInvocation") -> None:
        self._cache[invocation.invocation_id] = invocation

    def _get_invocation(
        self, invocation_id: str
    ) -> "DistributedInvocation[Params, Result]":
        return self._cache[invocation_id]

    def _add_history(
        self,
        invocation: "DistributedInvocation",
        invocation_history: "InvocationHistory",
    ) -> None:
        self._history[invocation.invocation_id].append(invocation_history)

    def _get_history(
        self, invocation: "DistributedInvocation[Params, Result]"
    ) -> list[InvocationHistory]:
        return self._history[invocation.invocation_id]

    def _set_result(
        self, invocation: "DistributedInvocation[Params, Result]", result: "Result"
    ) -> None:
        self._results[invocation.invocation_id] = result

    def _set_exception(
        self,
        invocation: "DistributedInvocation[Params, Result]",
        exception: "Exception",
    ) -> None:
        self._exceptions[invocation.invocation_id] = exception

    def _get_result(
        self, invocation: "DistributedInvocation[Params, Result]"
    ) -> "Result":
        return self._results[invocation.invocation_id]

    def _get_exception(
        self, invocation: "DistributedInvocation[Params, Result]"
    ) -> Exception:
        return self._exceptions[invocation.invocation_id]



from .config_base import ConfigBase, ConfigField


class ConfigRedis(ConfigBase):
    """Specific Configuration for any Redis client"""

    redis_host = ConfigField("localhost")
    redis_port = ConfigField(6379)
    redis_db = ConfigField(0)



from .config_base import ConfigBase, ConfigField, default_config_field_mapper
from .single_invocation_pending import (
    SingleInvocation,
    SingleInvocationPerArguments,
    SingleInvocationPerKeyArguments,
)

__all__ = [
    "ConfigBase",
    "ConfigField",
    "default_config_field_mapper",
    "SingleInvocation",
    "SingleInvocationPerArguments",
    "SingleInvocationPerKeyArguments",
]



from .config_base import ConfigBase, ConfigField


class ConfigRunner(ConfigBase):
    """Specific Configuration for any Runner"""

    invocation_wait_results_sleep_time_sec = ConfigField(0.1)
    runner_loop_sleep_time_sec = ConfigField(0.1)


class ConfigMemRunner(ConfigRunner):
    """Specific Configuration for the MemRunner"""

    invocation_wait_results_sleep_time_sec = ConfigField(0.01)
    runner_loop_sleep_time_sec = ConfigField(0.01)



import hashlib
import inspect
from functools import cached_property
from typing import TYPE_CHECKING, Any, Optional

if TYPE_CHECKING:
    from .types import Args, Func


class Arguments:
    def __init__(self, kwargs: Optional["Args"] = None) -> None:
        self.kwargs: "Args" = kwargs or {}

    @classmethod
    def from_call(cls, func: "Func", *args: Any, **kwargs: Any) -> "Arguments":
        sig = inspect.signature(func)
        bound_args = sig.bind(*args, **kwargs)
        bound_args.apply_defaults()
        return cls(bound_args.arguments)

    @cached_property
    def args_id(self) -> str:
        """Generate a unique id for these arguments"""
        if not self.kwargs:
            return "no_args"
        sorted_items = sorted(self.kwargs.items())
        args_str = "".join([f"{k}:{v}" for k, v in sorted_items])
        return hashlib.sha256(args_str.encode()).hexdigest()

    def __hash__(self) -> int:
        return hash(self.args_id)

    def __eq__(self, __value: object) -> bool:
        if not isinstance(__value, Arguments):
            return False
        return self.args_id == __value.args_id

    def __str__(self) -> str:
        return str(self.kwargs)

    def __repr__(self) -> str:
        return f"Arguments({self.kwargs})"



from .config_base import ConfigBase, ConfigField


class ConfigPynenc(ConfigBase):
    """
    Main config of the pynenc app.

    Attributes
    ----------
    dev_mode_force_sync_tasks : ConfigField
        If True, forces tasks to run synchronously, useful for development.
    max_pending_seconds : ConfigField
        Maximum time in seconds a task can remain in PENDING state before it expires.
        See :class:`~pynenc.invocation.status.InvocationStatus` for more details.
    logging_level : ConfigField
        The logging level of the application ('info', 'warning', 'error', etc.).
    """

    dev_mode_force_sync_tasks = ConfigField(False)
    max_pending_seconds = ConfigField(5.0)
    logging_level = ConfigField("info")



import multiprocessing
import threading
import time
from collections import defaultdict
from functools import cached_property
from typing import Any, NamedTuple, Optional

from pynenc.invocation import DistributedInvocation, InvocationStatus

from ..conf.config_runner import ConfigMemRunner
from .base_runner import BaseRunner


class ThreadInfo(NamedTuple):
    thread: threading.Thread
    invocation: DistributedInvocation


class MemRunner(BaseRunner):
    wait_conditions: dict["DistributedInvocation", threading.Condition]
    wait_invocation: dict["DistributedInvocation", set["DistributedInvocation"]]
    threads: dict[str, ThreadInfo]
    max_threads: int
    waiting_threads: int

    @cached_property
    def conf(self) -> ConfigMemRunner:
        return ConfigMemRunner(
            config_values=self.app.config_values,
            config_filepath=self.app.config_filepath,
        )

    def _on_start(self) -> None:
        # Initialize thread list and condition dictionary
        self.wait_conditions = defaultdict(threading.Condition)
        self.wait_invocation = defaultdict(set)
        self.threads = {}
        self.max_threads = multiprocessing.cpu_count()
        self.waiting_threads = 0

    def _on_stop(self) -> None:
        """kill all the running threads and change invocation status to retry"""
        for thread_info in self.threads.values():
            thread_info.thread.join()
            self.app.orchestrator.set_invocation_status(
                thread_info.invocation, InvocationStatus.RETRY
            )

    @property
    def available_threads(self) -> int:
        """Return the number of available threads"""
        self.threads = {k: v for k, v in self.threads.items() if v.thread.is_alive()}
        # do not consider waiting threads
        # in testing with only one thread available and only one worker
        # the waiting threads will prevent to run the invocation that will unlock
        # the invocations waiting in these threads
        return self.max_threads - len(self.threads)  # - self.waiting_threads

    def runner_loop_iteration(self) -> None:
        for invocation in self.app.orchestrator.get_invocations_to_run(
            max_num_invocations=self.available_threads
        ):
            thread = threading.Thread(target=invocation.run, daemon=True)
            thread.start()
            self.threads[invocation.invocation_id] = ThreadInfo(thread, invocation)
            self.logger.info(f"Running {invocation=} on {thread=}")
        for invocation in list(self.wait_conditions):
            if invocation.status.is_final():
                # Notify all waiting threads to continue
                with self.wait_conditions[invocation]:
                    self.wait_conditions[invocation].notify_all()
                self.wait_conditions.pop(invocation)
                # Set all waiting invocations status to running
                waiting_invocations = self.wait_invocation.pop(invocation)
                self.app.orchestrator.set_invocations_status(
                    list(waiting_invocations), InvocationStatus.RUNNING
                )
                self.waiting_threads -= len(waiting_invocations)
                self.logger.debug(
                    f"{invocation=} on final {invocation.status=}, resuming {waiting_invocations=}"
                )
        time.sleep(self.conf.runner_loop_sleep_time_sec)

    def waiting_for_results(
        self,
        running_invocation: Optional["DistributedInvocation"],
        result_invocations: list["DistributedInvocation"],
        runner_args: Optional[dict[str, Any]] = None,
    ) -> None:
        """In this case we let the current thread waiting in a condition based in the result invocation
        So we ignore the running_invocation parameter
        """
        del runner_args
        if not running_invocation:
            # running from outside this runner (user instantiate an app with this runner class,
            # but ask for an invocation result outside of the runner processes)
            self.logger.debug(
                f"Waiting for {result_invocations=} from outside this runner"
            )
            time.sleep(self.conf.invocation_wait_results_sleep_time_sec)
            return
        self.app.orchestrator.set_invocation_status(
            running_invocation, InvocationStatus.PAUSED
        )
        self.logger.debug(
            f"Pausing invocation {running_invocation.invocation_id} is waiting for others to finish"
        )
        for result_invocation in result_invocations:
            self.wait_invocation[result_invocation].add(running_invocation)
            self.waiting_threads += 1
            with self.wait_conditions[result_invocation]:
                self.logger.debug(
                    f"Invocation {running_invocation.invocation_id} is waiting for invocation {result_invocation.invocation_id} to finish"
                )
                self.wait_conditions[result_invocation].wait()



from .base_orchestrator import BaseOrchestrator
from .mem_orchestrator import MemOrchestrator
from .redis_orchestrator import RedisOrchestrator

__all__ = ["BaseOrchestrator", "MemOrchestrator", "RedisOrchestrator"]



import os
import signal
import socket
import threading
import time
import warnings
from abc import ABC, abstractmethod
from functools import cached_property
from typing import TYPE_CHECKING, Any, Optional

from pynenc.exceptions import RunnerNotExecutableError

from ..conf.config_runner import ConfigRunner
from ..util.log import RunnerLogAdapter

if TYPE_CHECKING:
    from types import FrameType

    from ..app import Pynenc
    from ..invocation import DistributedInvocation


class BaseRunner(ABC):
    """
    The Runner will execute invocations from the broker.

    - It requires an app because it needs to know about the broker, orchestrator, etc.
    - The runner will affect the behavior of the task result, for example:
      * In a subprocess environment, it may implement a pipe to communicate for pausing/resuming processes.
      * In an async environment, the value should be an async function to wait for distributed results.
      * In a cloud function environment aiming for speed with a single thread, it might not wait more than 'x' seconds and instead, create a 'callback', save the status, and convert the current execution into a task that will be called when the result is ready.
      * In a multiprocessing environment in a Kubernetes pod with capabilities to create new pods, it may have different behaviors.
      * For an asyncio worker, it runs several tasks in one processor, and the value should wait with async.
    """

    def __init__(self, app: "Pynenc") -> None:
        self.app = app
        self.app.runner = self
        self.running = False
        self.logger = RunnerLogAdapter(self.app.logger, self.runner_id)

    @cached_property
    def runner_id(self) -> str:
        hostname = socket.gethostname()
        pid = os.getpid()
        return f"{self.__class__.__name__}({hostname}-{pid})"

    @cached_property
    def conf(self) -> ConfigRunner:
        return ConfigRunner(
            config_values=self.app.config_values,
            config_filepath=self.app.config_filepath,
        )

    @abstractmethod
    def _on_start(self) -> None:
        """This method is called when the runner starts"""

    def on_start(self) -> None:
        """This method is called when the runner starts"""
        if threading.current_thread() is threading.main_thread():
            signal.signal(signal.SIGINT, self.stop_runner_loop)
            signal.signal(signal.SIGTERM, self.stop_runner_loop)
        else:
            warnings.warn(
                "Running in a secondary thread. Signal handling will be skipped.",
                stacklevel=2,
            )
        self.running = True
        self.logger.info("Starting runner...")
        self._on_start()

    @abstractmethod
    def _on_stop(self) -> None:
        """This method is called when the runner stops"""

    def on_stop(self) -> None:
        """This method is called when the runner stops"""
        self.running = False
        self.logger.info("Stopping runner...")
        self._on_stop()

    @abstractmethod
    def runner_loop_iteration(self) -> None:
        """
        One iteration of the runner loop.
        Subclasses should implement this method to process invocations.
        """

    def stop_runner_loop(
        self, signum: Optional[int] = None, frame: Optional["FrameType"] = None
    ) -> None:
        """Stops the runner loop"""
        self.app.logger.info(
            f"Received signal {signum=} {frame=} Stopping runner loop..."
        )
        self.running = False

    @abstractmethod
    def waiting_for_results(
        self,
        running_invocation: Optional["DistributedInvocation"],
        result_invocation: list["DistributedInvocation"],
        runner_args: Optional[dict[str, Any]] = None,
    ) -> None:
        """This method is called from the result method of an invocation
        It signals the runner that the running invocation is waiting for the result of the result invocation

        The running invocation may be None, when the result was called from outside a runner (e.g. user environment)
        In that case will be handle by the DummyRunner (default in the pynenc app to handle this cases)

        The runner has the oportunity to define the waiting behaviour of the running invocation in this method
        Otherwise the running invocation will infinetely loop until the result invocation is ready
        """

    def run(self) -> None:
        """Starts the runner"""
        self.on_start()
        try:
            while self.running:
                self.runner_loop_iteration()
        except KeyboardInterrupt:
            self.logger.warning("KeyboardInterrupt received. Stopping runner...")
        except Exception as e:
            self.logger.exception(f"Exception in runner loop: {e}")
            raise e
        finally:
            self.on_stop()


class DummyRunner(BaseRunner):
    """
    This runner is a placeholder for the Pynenc app.
    It will be used when the app is defined in any other Python environment than a Pynenc runner.

    Examples include:
      - A script that defines the app, decorates some tasks, routes them, and then finishes. Such a script does not plan to run anything itself but triggers tasks that will later run in actual runners.
    """

    def _on_start(self) -> None:
        raise RunnerNotExecutableError(
            "This runner is a placeholder for the Pynenc app"
        )

    def _on_stop(self) -> None:
        raise RunnerNotExecutableError(
            "This runner is a placeholder for the Pynenc app"
        )

    def runner_loop_iteration(self) -> None:
        raise RunnerNotExecutableError(
            "This runner is a placeholder for the Pynenc app"
        )

    def waiting_for_results(
        self,
        running_invocation: Optional["DistributedInvocation"],
        result_invocation: list["DistributedInvocation"],
        runner_args: Optional[dict[str, Any]] = None,
    ) -> None:
        del running_invocation, result_invocation, runner_args
        # invocation.result() was called from outside a runner
        # it will block and loop indefinetely until result is available
        time.sleep(self.conf.invocation_wait_results_sleep_time_sec)



import os
import signal
import time
from multiprocessing import Manager, Process, cpu_count
from typing import TYPE_CHECKING, Any, Optional

from pynenc.invocation import InvocationStatus

from ..exceptions import RunnerError
from .base_runner import BaseRunner

if TYPE_CHECKING:
    from ..invocation import DistributedInvocation


class ProcessRunner(BaseRunner):
    wait_invocation: dict["DistributedInvocation", set["DistributedInvocation"]]
    processes: dict["DistributedInvocation", Process]

    max_processes: int

    @property
    def runner_args(self) -> dict[str, Any]:
        return {"wait_invocation": self.wait_invocation}

    @property
    def waiting_processes(self) -> int:
        if not self.wait_invocation:
            return 0
        return len(set.union(*self.wait_invocation.values()))

    def parse_args(self, args: dict[str, Any]) -> None:
        self.wait_invocation = args["wait_invocation"]

    def _on_start(self) -> None:
        self.wait_invocation = Manager().dict()  # type: ignore
        self.processes = {}
        self.max_processes = cpu_count()

    def _on_stop(self) -> None:
        """kill all the running processes and change invocation status to retry"""
        for invocation, process in self.processes.items():
            process.kill()
            self.app.orchestrator.set_invocation_status(
                invocation, InvocationStatus.RETRY
            )

    @property
    def available_processes(self) -> int:
        for invocation in list(self.processes.keys()):
            if not self.processes[invocation].is_alive():
                del self.processes[invocation]
        # discount waiting processes, they should do nothing
        # until the blocking invocation is finished
        # otherwise, running one worker with one process
        # will be lock indefintely until the blocking invocation runs
        return self.max_processes - len(self.processes)  # - self.waiting_processes

    def runner_loop_iteration(self) -> None:
        # called from parent process memory space
        for invocation in self.app.orchestrator.get_invocations_to_run(
            max_num_invocations=self.available_processes
        ):
            process = Process(
                target=invocation.run,
                kwargs={"runner_args": self.runner_args},
                daemon=True,
            )
            process.start()
            self.logger.debug(
                f"Running invocation {invocation.invocation_id} on {process.pid=}"
            )
            if process.pid:
                self.processes[invocation] = process
            else:
                ...
                # TODO if for mypy, the process should have a pid after start, otherwise it should raise an exception

        for invocation in list(self.wait_invocation.keys()):
            is_final = invocation.status.is_final()
            for waiting_invocation in self.wait_invocation[invocation]:
                if pid := self.processes[waiting_invocation].pid:
                    if is_final:
                        os.kill(pid, signal.SIGCONT)
                    else:
                        os.kill(pid, signal.SIGSTOP)
            if is_final:
                waiting_invocations = self.wait_invocation.pop(invocation)
                self.logger.debug(
                    f"{invocation=} on final {invocation.status=}, resuming {waiting_invocations=}"
                )
                self.app.orchestrator.set_invocations_status(
                    list(waiting_invocations), InvocationStatus.RUNNING
                )
        time.sleep(self.conf.runner_loop_sleep_time_sec)

    def waiting_for_results(
        self,
        running_invocation: Optional["DistributedInvocation"],
        result_invocations: list["DistributedInvocation"],
        runner_args: Optional[dict[str, Any]] = None,
    ) -> None:
        # called from subprocess memory space
        if not running_invocation:
            time.sleep(self.conf.invocation_wait_results_sleep_time_sec)
            return

        self.app.orchestrator.set_invocation_status(
            running_invocation, InvocationStatus.PAUSED
        )
        if not result_invocations:
            return
        if not runner_args:
            raise RunnerError("runner_args should be defined for ProcessRunner")
        self.parse_args(runner_args)
        for result_invocation in result_invocations:
            if result_invocation not in self.wait_invocation:
                self.wait_invocation[result_invocation] = set()
            self.logger.debug(
                f"Invocation {running_invocation.invocation_id} is waiting for invocation {result_invocation.invocation_id} to finish"
            )
            self.wait_invocation[result_invocation].add(running_invocation)



import os
from collections import defaultdict
from typing import Any, Callable, Generic, Iterator, Optional, Type, TypeVar, cast

from ..exceptions import ConfigMultiInheritanceError
from ..util import files

T = TypeVar("T")

ConfigLoader = Callable[[str], dict[str, str]]
ConfigFieldMapper = Callable[[Any, Type[T]], T]


def default_config_field_mapper(value: Any, expected_type: Type[T]) -> T:
    if isinstance(value, expected_type):
        return value
    if callable(expected_type):
        try:
            callable_type = cast(Callable[[Any], T], expected_type)
            return callable_type(value)  # type conversion
        except (ValueError, TypeError) as ex:
            raise TypeError(
                f"Invalid type. Expected {expected_type} instead {type(value)}."
            ) from ex
    else:
        raise TypeError(f"Cannot convert to {expected_type}")


class ConfigField(Generic[T]):
    """
    Define each typed field from a ConfigBase instance.

    This class is used to define typed configuration fields within a ConfigBase
    subclass. It ensures type consistency and supports value validation and casting.

    Parameters
    ----------
    default_value : T
        The default value for the configuration field.
    mapper : Optional[ConfigFieldMapper]
        An optional function to map or transform the value.

    Attributes
    ----------
    _default_value : T
        Stores the default value of the configuration field.
    _mapper : ConfigFieldMapper
        The function used for mapping or transforming the value.
    """

    def __init__(
        self, default_value: T, mapper: Optional[ConfigFieldMapper] = None
    ) -> None:
        self._default_value: T = default_value
        self._mapper = mapper or default_config_field_mapper

    def __get__(self, instance: Optional["ConfigBase"], owner: Type[object]) -> T:
        del owner
        if instance is None:
            return self._default_value
        return instance._config_values.get(self, self._default_value)

    def __set__(self, instance: "ConfigBase", value: Any) -> None:
        instance._config_values[self] = self._mapper(value, type(self._default_value))


ENV_PREFIX = "PYNENC"
ENV_SEP = "__"
ENV_FILEPATH = "FILEPATH"


def get_env_key(field: str, config: Optional[Type["ConfigBase"]] = None) -> str:
    """gets the key used in the environment variables"""
    if config:
        return f"{ENV_PREFIX}{ENV_SEP}{config.__name__.upper()}{ENV_SEP}{field.upper()}"
    return f"{ENV_PREFIX}{ENV_SEP}{field.upper()}"


class ConfigBase:
    """
    Base class for defining configuration settings.

    This class serves as the base for creating configuration classes. It supports
    hierarchical and flexible configuration from various sources, including
    environment variables, configuration files, and default values.

    Configuration values are determined based on the following priority (highest to lowest):
    1. Direct assignment in the config instance (not recommended)
    2. Environment variables
    3. Configuration file path specified by environment variables
    4. Configuration file path (YAML, TOML, JSON) by config_filepath parameter
    5. `pyproject.toml`
    6. Default values specified in the `ConfigField`
    7. Previous steps for any Parent config class
    8. User does not specify anything (default values)

    Examples
    --------
    Define a configuration class for a Redis client:

    .. code-block:: python

        class ConfigRedis(ConfigBase):
            redis_host = ConfigField("localhost")
            redis_port = ConfigField(6379)
            redis_db = ConfigField(0)

    Define a main configuration class for orchestrator components:

    .. code-block:: python

        class ConfigOrchestrator(ConfigBase):
            cycle_control = ConfigField(True)
            blocking_control = ConfigField(True)
            auto_final_invocation_purge_hours = ConfigField(24.0)

    Combine configurations using multiple inheritance:

    .. code-block:: python

        class ConfigOrchestratorRedis(ConfigOrchestrator, ConfigRedis):
            pass

    The `ConfigOrchestratorRedis` class now includes settings from both `ConfigOrchestrator`
    and `ConfigRedis`.
    """

    def __init__(
        self,
        config_values: Optional[dict[str, Any]] = None,
        config_filepath: Optional[str] = None,
    ) -> None:
        self.config_cls_to_fields: dict[str, set[str]] = defaultdict(set)
        _ = avoid_multi_inheritance_field_conflict(
            self.__class__, self.config_cls_to_fields
        )
        self._config_values: dict[ConfigField, Any] = {}

        # on the first run, we load defaults values specified in the mapping
        # afterwards, that values will be modified by the ancestors
        # the childs will have higher priority
        self._mapped_keys: set[str] = set()
        self._mapped_environ_keys: set[str] = set()
        self.init_config_values(self.__class__, config_values, config_filepath)

    @staticmethod
    def get_config_id(config_cls: Type["ConfigBase"]) -> str:
        return config_cls.__name__.replace("Config", "").lower()

    def init_parent_values(
        self,
        config_cls: Type["ConfigBase"],
        config_values: Optional[dict[str, Any]],
        config_filepath: Optional[str],
    ) -> None:
        # Initialize parent classes that are subclasses of ConfigBase
        for parent in config_cls.__bases__:
            if issubclass(parent, ConfigBase) and parent is not ConfigBase:
                self.init_config_values(parent, config_values, config_filepath)

    def init_config_values(
        self,
        config_cls: Type["ConfigBase"],
        config_values: Optional[dict[str, Any]],
        config_filepath: Optional[str],
    ) -> None:
        config_id = self.get_config_id(config_cls)
        self.init_parent_values(config_cls, config_values, config_filepath)
        if not get_config_fields(self.__class__):
            raise TypeError(
                "Cannot instantiate a ConfigBase without any ConfigField attribute"
            )
        # 5.- User specifies the config by values (dict[str: Any])
        if config_values:
            self.init_config_value_from_mapping(
                "config_values", config_id, config_values
            )
        # 4.- User specifies config values in pyproject.toml
        if os.path.isfile("pyproject.toml"):
            self.init_config_value_from_mapping(
                "pyproject.toml", config_id, files.load_file("pyproject.toml")
            )
        # 3.- User specifies the config filepath(ref to a yml, toml or json…)
        if config_filepath:
            self.init_config_value_from_mapping(
                "config_filepath", config_id, files.load_file(config_filepath)
            )
        # 2.- User specifies the location of the config file by env vars
        # 2.1 Global config filepath specify by env var
        if filepath := os.environ.get(get_env_key(ENV_FILEPATH)):
            self.init_config_value_from_mapping(
                "ENV_FILEPATH", config_id, files.load_file(filepath)
            )
        # 2.2 Specific class config filepath specify by env var
        if filepath := os.environ.get(get_env_key(ENV_FILEPATH, config_cls)):
            self.init_config_value_from_mapping(
                "ENV_CLASS_FILEPATH", config_id, files.load_file(filepath)
            )
        # 1.- User specifies environment variables
        self.init_config_value_from_env_vars(config_cls)

    def init_config_value_from_mapping(
        self, source: str, config_id: str, mapping: dict[str, Any]
    ) -> None:
        conf_mapping = mapping.get(config_id, {})
        conf_mapping = conf_mapping if isinstance(conf_mapping, dict) else {}
        for key in self.config_cls_to_fields.get(self.__class__.__name__, []):
            general_key = f"{source}##{key}"
            class_key = f"{source}##{config_id}##{key}"
            if general_key not in self._mapped_keys and key in mapping:
                setattr(self, key, mapping[key])
                self._mapped_keys.add(general_key)
            if class_key not in self._mapped_keys and key in conf_mapping:
                setattr(self, key, conf_mapping[key])
                self._mapped_keys.add(class_key)

    def init_config_value_from_env_vars(self, config_cls: Type["ConfigBase"]) -> None:
        for key in self.config_cls_to_fields.get(config_cls.__name__, []):
            if get_env_key(key, config_cls) in os.environ:
                setattr(self, key, os.environ[get_env_key(key, config_cls)])
            elif get_env_key(key) in os.environ:
                setattr(self, key, os.environ[get_env_key(key)])


def get_config_fields(cls: Type) -> Iterator[str]:
    for key, value in cls.__dict__.items():
        if isinstance(value, ConfigField):
            yield key


def avoid_multi_inheritance_field_conflict(
    config_cls: Type, config_cls_to_fields: dict[str, set[str]]
) -> dict[str, str]:
    """
    Ensures that the same configuration field is not defined in multiple parent classes of a given configuration class.

    This function checks all parent classes of the provided configuration class that are subclasses of `ConfigBase`.
    It ensures that each configuration field is defined only once among all parent classes. If a field is found in
    multiple parent classes, a `ConfigMultiInheritanceError` is raised. This check ensures deterministic behavior
    in the configuration inheritance hierarchy.

    Parameters:
        config_cls (Type): The configuration class to check for field conflicts.

    Returns:
        dict[str, str]: A dictionary mapping each configuration field to the name of the parent class where it is defined.

    Raises:
        ConfigMultiInheritanceError: If a configuration field is found in multiple parent classes.

    Example:
        >>> class ParentConfig1(ConfigBase):
        ...     field1 = ConfigField(default_value=1)
        ...
        >>> class ParentConfig2(ConfigBase):
        ...     field2 = ConfigField(default_value=2)
        ...
        >>> class ChildConfig(ParentConfig1, ParentConfig2):
        ...     pass
        ...
        >>> avoid_multi_inheritance_field_conflict(ChildConfig)
        {'field1': 'ParentConfig1', 'field2': 'ParentConfig2'}
    """
    map_field_to_config_cls: dict[str, str] = {}
    cls_fields: set[str] = set()
    for parent in config_cls.__bases__:
        if not issubclass(parent, ConfigBase) or parent is ConfigBase:
            continue
        for key in get_config_fields(parent):
            if key in map_field_to_config_cls:
                raise ConfigMultiInheritanceError(
                    f"ConfigField {key} found in parent classes {parent.__name__} and {map_field_to_config_cls[key]}"
                )
            map_field_to_config_cls[key] = parent.__name__
            config_cls_to_fields[parent.__name__].add(key)
        # add current parent ancestor's fields that may not be specified in the current class
        map_field_to_config_cls.update(
            avoid_multi_inheritance_field_conflict(parent, config_cls_to_fields)
        )
        cls_fields = cls_fields.union(config_cls_to_fields[parent.__name__])
    config_cls_to_fields[config_cls.__name__] = cls_fields.union(
        set(get_config_fields(config_cls))
    )
    return map_field_to_config_cls


def get_config_family_fields(config_cls: Type) -> dict[str, str]:
    """"""
    config_fields: dict[str, str] = {}
    for key, value in config_cls.__dict__.items():
        if isinstance(value, ConfigField):
            if key in config_fields:
                raise ConfigMultiInheritanceError(
                    f"ConfigField {key} found in parent classes {config_cls.__name__} and {config_fields[key]}"
                )
            config_fields[key] = config_cls.__name__
    return config_fields



from functools import cached_property
from time import time
from typing import TYPE_CHECKING, Iterator, Optional

import redis

from ..call import Call
from ..conf.config_orchestrator import ConfigOrchestratorRedis
from ..exceptions import CycleDetectedError, PendingInvocationLockError
from ..invocation import DistributedInvocation, InvocationStatus
from ..types import Params, Result
from ..util.redis_keys import Key
from .base_orchestrator import BaseBlockingControl, BaseCycleControl, BaseOrchestrator

if TYPE_CHECKING:
    from ..app import Pynenc
    from ..task import Task


class RedisCycleControl(BaseCycleControl):
    """A directed acyclic graph representing the call dependencies"""

    def __init__(self, app: "Pynenc", client: redis.Redis) -> None:
        self.app = app
        self.key = Key(app.app_id, "cycle_control")
        self.client = client

    def purge(self) -> None:
        self.key.purge(self.client)

    def add_call_and_check_cycles(
        self, caller: "DistributedInvocation", callee: "DistributedInvocation"
    ) -> None:
        """
        Add a new invocation to the graph. This represents a dependency where the caller
        is dependent on the callee.

        Raises a CycleDetectedError if the invocation would cause a cycle.
        """
        if caller.call_id == callee.call_id:
            raise CycleDetectedError.from_cycle([caller.call])
        if cycle := self.find_cycle_caused_by_new_invocation(caller, callee):
            raise CycleDetectedError.from_cycle(cycle)
        self.client.set(self.key.invocation(caller.invocation_id), caller.to_json())
        self.client.set(self.key.invocation(callee.invocation_id), callee.to_json())
        self.client.sadd(
            self.key.call_to_invocation(caller.call_id), caller.invocation_id
        )
        self.client.sadd(
            self.key.call_to_invocation(callee.call_id), callee.invocation_id
        )
        self.client.set(self.key.call(caller.call_id), caller.call.to_json())
        self.client.set(self.key.call(callee.call_id), callee.call.to_json())
        # self.calls[caller.call_id] = caller.call # todo check if it work with invocation_id
        # self.calls[callee.call_id] = callee.call
        # self.call_to_invocation[caller.call_id][caller.invocation_id] = caller
        # self.call_to_invocation[callee.call_id][callee.invocation_id] = callee
        # self.edges[caller.call_id].add(callee.call_id)
        self.client.sadd(self.key.edge(caller.call_id), callee.call_id)

    def remove_edges(self, call_id: str) -> None:
        """Remove all edges from a call"""
        callee_calls = self.client.smembers(self.key.edge(call_id))
        self.client.delete(self.key.edge(call_id))
        for callee_call_id in callee_calls:
            self.remove_edges(callee_call_id)

    def clean_up_invocation_cycles(self, invocation: "DistributedInvocation") -> None:
        """
        Remove an invocation from the graph. Also removes any edges to or from the invocation.
        """
        self.client.delete(self.key.invocation(invocation.invocation_id))
        self.client.srem(
            self.key.call_to_invocation(invocation.call_id), invocation.invocation_id
        )
        remaining_call_invocations = self.client.smembers(
            self.key.call_to_invocation(invocation.call_id)
        )
        if not remaining_call_invocations:
            self.client.delete(self.key.call(invocation.call_id))
            self.client.delete(self.key.call_to_invocation(invocation.call_id))
            self.remove_edges(invocation.call_id)

    def find_cycle_caused_by_new_invocation(
        self, caller: "DistributedInvocation", callee: "DistributedInvocation"
    ) -> list["Call"]:
        """
        Determines if adding an edge from the caller to the callee would create a cycle.

        Parameters
        ----------
        caller : DistributedInvocation
            The invocation making the call.
        callee : DistributedInvocation
            The invocation being called.

        Returns
        -------
        list
            List of invocations that would form the cycle after adding the new invocation, else an empty list.
        """
        # Temporarily add the edge to check if it would cause a cycle
        self.client.sadd(self.key.edge(caller.call_id), callee.call_id)

        # Set for tracking visited nodes
        visited: set[str] = set()

        # List for tracking the nodes on the path from caller to callee
        path: list[str] = []

        cycle = self._is_cyclic_util(caller.call_id, visited, path)

        # Remove the temporarily added edge
        self.client.srem(self.key.edge(caller.call_id), callee.call_id)

        return cycle

    def _is_cyclic_util(
        self,
        current_call_id: str,
        visited: set[str],
        path: list[str],
    ) -> list["Call"]:
        visited.add(current_call_id)
        path.append(current_call_id)

        call_cycle = []
        for _neighbour_call_id in self.client.smembers(self.key.edge(current_call_id)):
            neighbour_call_id = _neighbour_call_id.decode()
            if neighbour_call_id not in visited:
                cycle = self._is_cyclic_util(neighbour_call_id, visited, path)
                if cycle:
                    return cycle
            elif neighbour_call_id in path:
                cycle_start_index = path.index(neighbour_call_id)
                for _id in path[cycle_start_index:]:
                    if call_json := self.client.get(self.key.call(_id)):
                        call_cycle.append(Call.from_json(self.app, call_json.decode()))
        path.pop()
        return call_cycle


class RedisBlockingControl(BaseBlockingControl):
    """A directed acyclic graph representing the call dependencies"""

    def __init__(self, app: "Pynenc", client: redis.Redis) -> None:
        self.app = app
        self.key = Key(app.app_id, "blocking_control")
        self.client = client

    def purge(self) -> None:
        self.key.purge(self.client)

    def waiting_for_results(
        self, waiter: "DistributedInvocation", waiteds: list["DistributedInvocation"]
    ) -> None:
        """
        Register that an invocation (waiter) is waiting for the results of another invocation (waited).
        """
        waited_invocation_ids = []
        for waited in waiteds:
            waited_invocation_ids.append(waited.invocation_id)
            self.client.set(self.key.invocation(waited.invocation_id), waited.to_json())
            self.client.sadd(
                self.key.waited_by(waited.invocation_id), waiter.invocation_id
            )
            # Add the waited invocation to the 'all_waited' sorted set with the current time as the score
            self.client.zadd(self.key.all_waited(), {waited.invocation_id: time()})
            # If the waited invocation is not waiting for anything else, add it to the 'not_waiting' sorted set
            if not self.client.exists(self.key.waiting_for(waited.invocation_id)):
                self.client.zadd(self.key.not_waiting(), {waited.invocation_id: time()})
        # If the waiter is in the 'not_waiting' sorted set, remove it
        if self.client.zscore(self.key.not_waiting(), waiter.invocation_id) is not None:
            self.client.zrem(self.key.not_waiting(), waiter.invocation_id)
        self.client.sadd(
            self.key.waiting_for(waiter.invocation_id), *waited_invocation_ids
        )

    def release_waiters(self, invocation: "DistributedInvocation") -> None:
        """
        Remove an invocation from the graph. Also removes any edges to or from the invocation.
        """
        # for each invocation thas is waiting for the invocation
        for waited_invocation_id in self.client.smembers(
            self.key.waited_by(invocation.invocation_id)
        ):
            # remove the invocation from the list of invocations waited by the waiter
            self.client.srem(
                self.key.waiting_for(waited_invocation_id.decode()),
                invocation.invocation_id,
            )
            # if the waiter is not waiting for anything else, add it to the 'not_waiting' sorted set
            if not self.client.exists(self.key.waiting_for(waited_invocation_id)):
                self.client.zadd(self.key.not_waiting(), {waited_invocation_id: time()})
        self.client.delete(self.key.invocation(invocation.invocation_id))
        self.client.delete(self.key.waiting_for(invocation.invocation_id))
        self.client.delete(self.key.waited_by(invocation.invocation_id))
        self.client.zrem(self.key.all_waited(), invocation.invocation_id)
        self.client.zrem(self.key.not_waiting(), invocation.invocation_id)

    def get_blocking_invocations(
        self, max_num_invocations: int
    ) -> Iterator["DistributedInvocation[Params, Result]"]:
        """
        Returns the invocations that are blocking others but not waiting for anything themselves.
        The oldest invocations are returned first.

        Parameters
        ----------
        max_num_invocations : int
            The maximum number of invocations to return.

        Returns
        -------
        list[DistributedInvocation]
            A list of blocking invocations or an empty list if no invocations are blocking.
        """
        index = 0
        page_size = max(10, max_num_invocations)  # adjust as needed
        while max_num_invocations > 0:
            if not (
                page := self.client.zrange(
                    self.key.not_waiting(), index, index + page_size - 1
                )
            ):
                break
            index += page_size
            for waited_invocation_id in page:
                invocation_id = waited_invocation_id.decode()
                if inv := self.client.get(self.key.invocation(invocation_id)):
                    invocation = DistributedInvocation.from_json(self.app, inv.decode())
                    if self.app.orchestrator.get_invocation_status(
                        invocation
                    ).is_available_for_run():
                        max_num_invocations -= 1
                        yield invocation


class StatusNotFound(Exception):
    """Raised when a status is not found in Redis"""


class TaskRedisCache:
    def __init__(self, app: "Pynenc", client: redis.Redis) -> None:
        self.app = app
        self.client = client
        self.key = Key(app.app_id, "orchestrator")

    def purge(self) -> None:
        self.key.purge(self.client)

    def set_status(
        self,
        invocation: "DistributedInvocation[Params, Result]",
        status: "InvocationStatus",
    ) -> None:
        task_id = invocation.task.task_id
        invocation_id = invocation.invocation_id
        try:
            previous_status = self._get_invocation_status(invocation.invocation_id)
            # already exists in Redis, remove from previous status
            self.client.srem(self.key.status(task_id, previous_status), invocation_id)
        except StatusNotFound:
            # new invocation, init invocation in Redis
            for arg, val in invocation.serialized_arguments.items():
                self.client.sadd(self.key.args(task_id, arg, val), invocation_id)
            self.client.set(self.key.invocation(invocation_id), invocation.to_json())
            self.client.sadd(self.key.task(task_id), invocation_id)
        self.client.sadd(self.key.status(task_id, status), invocation_id)
        self.client.set(self.key.invocation_status(invocation_id), status.value)
        if status != InvocationStatus.PENDING:
            self.clean_pending_status(invocation)

    def set_pending_status(
        self, invocation: "DistributedInvocation[Params, Result]"
    ) -> None:
        invocation_id = invocation.invocation_id
        lock = self.client.lock(
            f"lock:pending_status:{invocation_id}",
            blocking_timeout=self.app.conf.max_pending_seconds,
        )
        if not lock.acquire(blocking=True):
            raise PendingInvocationLockError(invocation_id)
        try:
            self.client.set(self.key.pending_timer(invocation_id), time())
            previous_status = self.get_invocation_status(invocation)
            if previous_status == InvocationStatus.PENDING:
                raise PendingInvocationLockError(invocation_id)
            self.client.set(
                self.key.previous_status(invocation_id), previous_status.value
            )
            self.set_status(invocation, InvocationStatus.PENDING)
        finally:
            lock.release()

    def set_up_invocation_auto_purge(
        self, invocation: "DistributedInvocation[Params, Result]"
    ) -> None:
        self.client.zadd(
            self.key.invocation_auto_purge(),
            {invocation.invocation_id: time()},
        )

    def auto_purge(self) -> None:
        end_time = (
            time() - self.app.orchestrator.conf.auto_final_invocation_purge_hours * 3600
        )
        for _invocation_id in self.client.zrangebyscore(
            self.key.invocation_auto_purge(), 0, end_time
        ):
            invocation_id = _invocation_id.decode()
            if inv := self.client.get(self.key.invocation(invocation_id)):
                invocation = DistributedInvocation.from_json(self.app, inv.decode())
                self.client.delete(self.key.invocation(invocation_id))
                task_id = invocation.task.task_id
                # clean up task keys
                self.client.srem(self.key.task(task_id), invocation_id)
                if not self.client.smembers(self.key.task(task_id)):
                    self.client.delete(self.key.task(task_id))
                # clean up task-status keys
                self.client.srem(
                    self.key.status(task_id, invocation.status), invocation_id
                )
                if not self.client.smembers(
                    self.key.status(task_id, invocation.status)
                ):
                    self.client.delete(self.key.status(task_id, invocation.status))
            self.client.delete(self.key.invocation_status(invocation_id))
            self.client.zrem(self.key.invocation_auto_purge(), invocation_id)
            self.client.delete(self.key.pending_timer(invocation_id))
            self.client.delete(self.key.previous_status(invocation_id))

    def clean_pending_status(
        self, invocation: "DistributedInvocation[Params, Result]"
    ) -> None:
        self.client.delete(self.key.pending_timer(invocation.invocation_id))
        self.client.delete(self.key.previous_status(invocation.invocation_id))

    def _get_invocation_status(self, invocation_id: str) -> "InvocationStatus":
        if encoded_status := self.client.get(self.key.invocation_status(invocation_id)):
            return InvocationStatus(encoded_status.decode())
        raise StatusNotFound(f"Invocation status {invocation_id} not found in Redis")

    def get_invocation_status(
        self, invocation: "DistributedInvocation"
    ) -> "InvocationStatus":
        status = self._get_invocation_status(invocation.invocation_id)
        if status == InvocationStatus.PENDING:
            if encoded_pending_timer := self.client.get(
                self.key.pending_timer(invocation.invocation_id)
            ):
                elapsed = time() - float(encoded_pending_timer.decode())
                if elapsed > self.app.conf.max_pending_seconds:
                    if encoded_previous_status := self.client.get(
                        self.key.previous_status(invocation.invocation_id)
                    ):
                        previous_status = InvocationStatus(
                            encoded_previous_status.decode()
                        )
                        self.set_status(invocation, previous_status)
                        return previous_status
        return status

    def get_invocations(
        self,
        task_id: str,
        key_arguments: Optional[dict[str, str]],
        status: Optional["InvocationStatus"],
    ) -> Iterator["DistributedInvocation"]:
        # Start with the set of obj_ids for the task_id
        invocation_ids = self.client.smembers(self.key.task(task_id))

        # If key_arguments were provided, intersect the current obj_ids with those matching each arg:val pair
        if key_arguments:
            for arg, val in key_arguments.items():
                arg_val_ids = self.client.smembers(self.key.args(task_id, arg, val))
                invocation_ids = invocation_ids.intersection(arg_val_ids)

        # If status was provided, intersect the current obj_ids with those matching the status
        if status:
            status_ids = self.client.smembers(self.key.status(task_id, status))
            invocation_ids = invocation_ids.intersection(status_ids)

        # Now obj_ids contains only the ids of obj_strings that match all the provided keys
        # Fetch the obj_strings for these ids
        for invocation_id in invocation_ids:
            if inv := self.client.get(self.key.invocation(invocation_id.decode())):
                yield DistributedInvocation.from_json(self.app, inv.decode())


class RedisOrchestrator(BaseOrchestrator):
    def __init__(self, app: "Pynenc") -> None:
        super().__init__(app)
        self.client = redis.Redis(
            host=self.conf.redis_host, port=self.conf.redis_port, db=self.conf.redis_db
        )
        self.redis_cache = TaskRedisCache(app, self.client)
        self._cycle_control: Optional[RedisCycleControl] = None
        self._blocking_control: Optional[RedisBlockingControl] = None

    @cached_property
    def conf(self) -> ConfigOrchestratorRedis:
        return ConfigOrchestratorRedis(
            config_values=self.app.config_values,
            config_filepath=self.app.config_filepath,
        )

    @property
    def cycle_control(self) -> "RedisCycleControl":
        if not self._cycle_control:
            self._cycle_control = RedisCycleControl(self.app, self.client)
        return self._cycle_control

    @property
    def blocking_control(self) -> "RedisBlockingControl":
        if not self._blocking_control:
            self._blocking_control = RedisBlockingControl(self.app, self.client)
        return self._blocking_control

    def get_existing_invocations(
        self,
        task: "Task[Params, Result]",
        key_serialized_arguments: Optional[dict[str, str]] = None,
        status: Optional["InvocationStatus"] = None,
    ) -> Iterator["DistributedInvocation"]:
        return self.redis_cache.get_invocations(
            task.task_id, key_serialized_arguments, status
        )

    def _set_invocation_status(
        self,
        invocation: "DistributedInvocation[Params, Result]",
        status: "InvocationStatus",
    ) -> None:
        self.redis_cache.set_status(invocation, status)
        self.app.logger.debug(f"Set status of invocation {invocation} to {status}")

    def _set_invocation_pending_status(
        self, invocation: "DistributedInvocation"
    ) -> None:
        self.redis_cache.set_pending_status(invocation)
        self.app.logger.debug(f"Set status of invocation {invocation} to pending")

    def set_up_invocation_auto_purge(
        self, invocation: DistributedInvocation[Params, Result]
    ) -> None:
        self.redis_cache.set_up_invocation_auto_purge(invocation)

    def auto_purge(self) -> None:
        self.redis_cache.auto_purge()

    def get_invocation_status(
        self, invocation: "DistributedInvocation[Params, Result]"
    ) -> "InvocationStatus":
        return self.redis_cache.get_invocation_status(invocation)

    def purge(self) -> None:
        """Remove all invocations from the orchestrator"""
        self.redis_cache.purge()
        self.cycle_control.purge()
        self.blocking_control.purge()



import pickle
import threading
from collections import OrderedDict, defaultdict, deque
from time import time
from typing import TYPE_CHECKING, Any, Generic, Iterator

from ..exceptions import CycleDetectedError, PendingInvocationLockError
from ..invocation import DistributedInvocation, InvocationStatus
from ..types import Params, Result
from .base_orchestrator import BaseBlockingControl, BaseCycleControl, BaseOrchestrator

if TYPE_CHECKING:
    from ..app import Pynenc
    from ..call import Call
    from ..task import Task


class MemCycleControl(BaseCycleControl):
    """A directed acyclic graph representing the call dependencies"""

    def __init__(self, app: "Pynenc") -> None:
        self.app = app
        self.invocations: dict[str, DistributedInvocation] = {}
        self.calls: dict[str, "Call"] = {}
        self.call_to_invocation: dict[
            str, OrderedDict[str, DistributedInvocation]
        ] = defaultdict(OrderedDict)
        self.edges: dict[str, set[str]] = defaultdict(set)

    def add_call_and_check_cycles(
        self, caller: DistributedInvocation, callee: DistributedInvocation
    ) -> None:
        """
        Add a new invocation to the graph. This represents a dependency where the caller
        is dependent on the callee.

        Raises a CycleDetectedError if the invocation would cause a cycle.
        """
        if caller.call_id == callee.call_id:
            raise CycleDetectedError.from_cycle([caller.call])
        if cycle := self.find_cycle_caused_by_new_invocation(caller, callee):
            raise CycleDetectedError.from_cycle(cycle)
        self.invocations[caller.invocation_id] = caller
        self.invocations[callee.invocation_id] = callee
        self.calls[caller.call_id] = caller.call
        self.calls[callee.call_id] = callee.call
        self.call_to_invocation[caller.call_id][caller.invocation_id] = caller
        self.call_to_invocation[callee.call_id][callee.invocation_id] = callee
        self.edges[caller.call_id].add(callee.call_id)

    def clean_up_invocation_cycles(self, invocation: DistributedInvocation) -> None:
        """
        Remove an invocation from the graph. Also removes any edges to or from the invocation.
        """
        call_id = invocation.call_id
        if call_id in self.call_to_invocation:
            self.call_to_invocation[call_id].pop(invocation.invocation_id, None)
            if not self.call_to_invocation[call_id]:
                del self.call_to_invocation[call_id]
                if call_id in self.edges:
                    del self.edges[call_id]
                for edges in self.edges.values():
                    edges.discard(call_id)

    def find_cycle_caused_by_new_invocation(
        self, caller: DistributedInvocation, callee: DistributedInvocation
    ) -> list["Call"]:
        """
        Determines if adding an edge from the caller to the callee would create a cycle.

        Parameters
        ----------
        caller : DistributedInvocation
            The invocation making the call.
        callee : DistributedInvocation
            The invocation being called.

        Returns
        -------
        list
            List of invocations that would form the cycle after adding the new invocation, else an empty list.
        """
        # Temporarily add the edge to check if it would cause a cycle
        self.edges[caller.call_id].add(callee.call_id)

        # Set for tracking visited nodes
        visited: set[str] = set()

        # List for tracking the nodes on the path from caller to callee
        path: list[str] = []

        cycle = self._is_cyclic_util(caller.call_id, visited, path)

        # Remove the temporarily added edge
        self.edges[caller.call_id].remove(callee.call_id)

        return cycle

    def _is_cyclic_util(
        self,
        current_call_id: str,
        visited: set[str],
        path: list[str],
    ) -> list["Call"]:
        visited.add(current_call_id)
        path.append(current_call_id)

        for neighbour_call_id in self.edges.get(current_call_id, []):
            if neighbour_call_id not in visited:
                cycle = self._is_cyclic_util(neighbour_call_id, visited, path)
                if cycle:
                    return cycle
            elif neighbour_call_id in path:
                cycle_start_index = path.index(neighbour_call_id)
                return [self.calls[_id] for _id in path[cycle_start_index:]]

        path.pop()
        return []


class MemBlockingControl(BaseBlockingControl):
    """A directed acyclic graph representing the call dependencies"""

    def __init__(self, app: "Pynenc") -> None:
        self.app = app
        self.invocations: dict[str, DistributedInvocation] = {}
        self.waiting_for: dict[str, set[str]] = defaultdict(set)
        self.waited_by: dict[str, set[str]] = OrderedDict()

    def waiting_for_results(
        self,
        caller_invocation: "DistributedInvocation[Params, Result]",
        result_invocations: list["DistributedInvocation[Params, Result]"],
    ) -> None:
        """
        Register that an invocation (waiter) is waiting for the results of another invocation (waited).
        """
        waiter = caller_invocation
        for waited in result_invocations:
            self.invocations[waited.invocation_id] = waited
            self.waiting_for[waiter.invocation_id].add(waited.invocation_id)
            if waited.invocation_id not in self.waited_by:
                self.waited_by[waited.invocation_id] = set()
            self.waited_by[waited.invocation_id].add(waiter.invocation_id)

    def release_waiters(self, invocation: DistributedInvocation) -> None:
        """
        Remove an invocation from the graph. Also removes any edges to or from the invocation.
        """
        for waiter_id in self.waited_by.get(invocation.invocation_id, []):
            self.waiting_for[waiter_id].discard(invocation.invocation_id)
            if not self.waiting_for[waiter_id]:
                del self.waiting_for[waiter_id]
        self.waited_by.pop(invocation.invocation_id, None)
        self.waiting_for.pop(invocation.invocation_id, None)

    def get_blocking_invocations(
        self, max_num_invocations: int
    ) -> Iterator["DistributedInvocation[Params, Result]"]:
        """
        Returns the invocations that are blocking others but not waiting for anything themselves.
        The oldest invocations are returned first.

        Parameters
        ----------
        max_num_invocations : int
            The maximum number of invocations to return.

        Returns
        -------
        Set[DistributedInvocation] | None
            A set of blocking invocations or None if no invocations are blocking.
        """
        for inv_id in self.waited_by:
            if inv_id not in self.waiting_for:
                if self.app.orchestrator.get_invocation_status(
                    self.invocations[inv_id]
                ).is_available_for_run():
                    max_num_invocations -= 1
                    yield self.invocations[inv_id]
                    if max_num_invocations == 0:
                        return


class ArgPair:
    """Helper to simulate a Memory cache for key:value pairs in Task Invocations"""

    def __init__(self, key: str, value: Any) -> None:
        self.key = key
        self.value = value

    def __hash__(self) -> int:
        return hash((self.key, pickle.dumps(self.value)))

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, ArgPair):
            return self.key == other.key and self.value == other.value
        return False

    def __str__(self) -> str:
        return f"{self.key}:{self.value}"

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.__str__()})"


class TaskInvocationCache(Generic[Result]):
    def __init__(self, app: "Pynenc") -> None:
        self.app = app
        self.invocations: dict[str, DistributedInvocation] = {}
        self.args_index: dict[ArgPair, set[str]] = defaultdict(set)
        self.status_index: dict[InvocationStatus, set[str]] = defaultdict(set)
        self.pending_timer: dict[str, float] = {}
        self.pre_pending_status: dict[str, InvocationStatus] = {}
        self.invocation_status: dict[str, InvocationStatus] = {}
        self.invocations_to_purge: deque[tuple[float, str]] = deque()
        self.locks: dict[str, threading.Lock] = {}

    def get_invocations(
        self,
        key_arguments: dict[str, str] | None,
        status: InvocationStatus | None,
    ) -> Iterator[DistributedInvocation]:
        # Filtering by key_arguments
        if key_arguments:
            matches: list[set[str]] = []
            # Check all the task invocations related per each key_argument
            for key, value in key_arguments.items():
                invocation_ids = set()
                for _id in self.args_index[ArgPair(key, value)]:
                    # Check if the invocation also matches the (optionally) specified status
                    if not status or _id in self.status_index[status]:
                        invocation_ids.add(_id)
                # add the matching invocations for the key:val pair to the list
                if invocation_ids:
                    matches.append(invocation_ids)
            # yield the invocations that matches all the specified key:val pairs
            if matches:
                for invocation_id in set.intersection(*matches):
                    yield self.invocations[invocation_id]
        # Filtered only by statys
        elif status:
            for invocation_id in self.status_index[status]:
                yield self.invocations[invocation_id]
        # No filters, return all invocations
        else:
            yield from self.invocations.values()

    def set_up_invocation_auto_purge(
        self, invocation: "DistributedInvocation[Params, Result]"
    ) -> None:
        self.invocations_to_purge.append((time(), invocation.invocation_id))

    def auto_purge(self) -> None:
        end_time = (
            time() - self.app.orchestrator.conf.auto_final_invocation_purge_hours * 3600
        )
        while self.invocations_to_purge and self.invocations_to_purge[0][0] <= end_time:
            _, elem = self.invocations_to_purge.popleft()
            self.clean_up_invocation(elem)

    def clean_up_invocation(self, invocation_id: str) -> None:
        invocation = self.invocations.pop(invocation_id, None)
        if invocation:
            for key, value in invocation.serialized_arguments.items():
                self.args_index[ArgPair(key, value)].discard(invocation_id)
            self.status_index[self.invocation_status[invocation_id]].discard(
                invocation_id
            )
            self.invocation_status.pop(invocation_id, None)
            self.pending_timer.pop(invocation_id, None)
            self.pre_pending_status.pop(invocation_id, None)

    def set_status(
        self,
        invocation: "DistributedInvocation[Params, Result]",
        status: InvocationStatus,
    ) -> None:
        if status != InvocationStatus.PENDING:
            self.clean_pending_status(invocation)
        if (_id := invocation.invocation_id) not in self.invocations:
            self.invocations[_id] = invocation
            for key, value in invocation.serialized_arguments.items():
                self.args_index[ArgPair(key, value)].add(_id)
            self.status_index[status].add(_id)
        else:
            # already exists, remove previous status
            self.status_index[self.invocation_status[_id]].discard(_id)
            self.status_index[status].add(_id)
        self.invocation_status[_id] = status

    def clean_pending_status(
        self, invocation: "DistributedInvocation[Params, Result]"
    ) -> None:
        self.pending_timer.pop(invocation.invocation_id, None)
        self.pre_pending_status.pop(invocation.invocation_id, None)

    def set_pending_status(
        self, invocation: "DistributedInvocation[Params, Result]"
    ) -> None:
        invocation_id = invocation.invocation_id
        lock = self.locks.setdefault(invocation_id, threading.Lock())
        if not lock.acquire(False):
            raise PendingInvocationLockError(invocation_id)
        try:
            self.pending_timer[invocation_id] = time()
            previous_status = self.invocation_status[invocation.invocation_id]
            if previous_status == InvocationStatus.PENDING:
                raise PendingInvocationLockError(invocation_id)
            self.pre_pending_status[invocation_id] = previous_status
            self.set_status(invocation, InvocationStatus.PENDING)
        finally:
            lock.release()

    def get_status(
        self, invocation: "DistributedInvocation[Params, Result]"
    ) -> InvocationStatus:
        status = self.invocation_status[invocation.invocation_id]
        if status == InvocationStatus.PENDING:
            elapsed = time() - self.pending_timer[invocation.invocation_id]
            if elapsed > self.app.conf.max_pending_seconds:
                pre_pending_status = self.pre_pending_status[invocation.invocation_id]
                self.set_status(invocation, pre_pending_status)
                return pre_pending_status
        return status


class MemOrchestrator(BaseOrchestrator):
    def __init__(self, app: "Pynenc") -> None:
        self.cache: dict[str, TaskInvocationCache] = defaultdict(
            lambda: TaskInvocationCache(app)
        )
        self._cycle_control: MemCycleControl | None = None
        self._blocking_control: MemBlockingControl | None = None
        super().__init__(app)

    @property
    def cycle_control(self) -> "MemCycleControl":
        if not self._cycle_control:
            self._cycle_control = MemCycleControl(self.app)
        return self._cycle_control

    @property
    def blocking_control(self) -> "MemBlockingControl":
        if not self._blocking_control:
            self._blocking_control = MemBlockingControl(self.app)
        return self._blocking_control

    def get_existing_invocations(
        self,
        task: "Task[Params, Result]",
        key_serialized_arguments: dict[str, str] | None = None,
        status: InvocationStatus | None = None,
    ) -> Iterator[DistributedInvocation]:
        return self.cache[task.task_id].get_invocations(
            key_serialized_arguments, status
        )

    def _set_invocation_status(
        self,
        invocation: "DistributedInvocation[Params, Result]",
        status: InvocationStatus,
    ) -> None:
        self.cache[invocation.task.task_id].set_status(invocation, status)

    def _set_invocation_pending_status(
        self, invocation: DistributedInvocation[Params, Result]
    ) -> None:
        self.cache[invocation.task.task_id].set_pending_status(invocation)

    def set_up_invocation_auto_purge(
        self, invocation: DistributedInvocation[Params, Result]
    ) -> None:
        self.cache[invocation.task.task_id].set_up_invocation_auto_purge(invocation)

    def auto_purge(self) -> None:
        for cache in self.cache.values():
            cache.auto_purge()

    def get_invocation_status(
        self, invocation: "DistributedInvocation[Params, Result]"
    ) -> InvocationStatus:
        return self.cache[invocation.task.task_id].get_status(invocation)

    def purge(self) -> None:
        self.cache.clear()
        self._cycle_control = None
        self._blocking_control = None



from typing import TYPE_CHECKING, Any, Optional

if TYPE_CHECKING:
    from .invocation.dist_invocation import DistributedInvocation

# invocation_context keeps the current invocation, so it can be referenced as a parent from any sub-invocation
invocation_context: dict[str, Optional["DistributedInvocation"]] = {}
# runner_args keeps the arguments passed from the runner to a distributed invocation:
# - this is necessary for parent-subprocess communication on ProcessRunner
runner_args: Optional[dict[str, Any]] = None



from __future__ import annotations

import uuid
from abc import ABC, abstractmethod
from dataclasses import dataclass
from functools import cached_property
from typing import TYPE_CHECKING, Any, Generic, Iterator, TypeVar

from ..call import Call
from ..types import Params, Result
from ..util.log import TaskLoggerAdapter

if TYPE_CHECKING:
    from ..app import Pynenc
    from ..arguments import Arguments
    from ..task import Task


T = TypeVar("T", bound="BaseInvocation")


@dataclass(frozen=True)
class BaseInvocation(ABC, Generic[Params, Result]):
    """Invocation of a task call

    A call can have several invocations in the system"""

    call: Call[Params, Result]

    def __post_init__(self) -> None:
        self.task.logger = TaskLoggerAdapter(
            self.app.logger, self.task.task_id, self.invocation_id
        )

    @property
    def app(self) -> Pynenc:
        return self.call.app

    @property
    def task(self) -> Task[Params, Result]:
        return self.call.task

    @property
    def arguments(self) -> Arguments:
        return self.call.arguments

    @property
    def serialized_arguments(self) -> dict[str, str]:
        return self.call.serialized_arguments

    @abstractmethod
    def to_json(self) -> str:
        """Returns a string with the serialized invocation"""

    @classmethod
    @abstractmethod
    def from_json(cls: type[T], app: Pynenc, serialized: str) -> T:
        """Returns a new invocation from a serialized invocation"""

    @property
    def call_id(self) -> str:
        return self.call.call_id

    @cached_property
    def invocation_id(self) -> str:
        """Returns a unique id for this invocation

        A task with the same arguments can have multiple invocations, the invocation id is used to differentiate them
        """
        return str(uuid.uuid4())

    @property
    @abstractmethod
    def result(self) -> Result:
        """"""

    def __str__(self) -> str:
        return f"{self.__class__.__name__}(invocation_id={self.invocation_id}, {self.call})"

    def __repr__(self) -> str:
        return self.__str__()

    def __hash__(self) -> int:
        return hash(self.invocation_id)

    def __eq__(self, other: Any) -> bool:
        # TODO equality based in task and arguments or in invocation_id?
        if not isinstance(other, BaseInvocation):
            return False
        return self.invocation_id == other.invocation_id


@dataclass(frozen=True)
class BaseInvocationGroup(ABC, Generic[Params, Result, T]):
    task: Task
    invocations: list[T]

    @property
    def app(self) -> Pynenc:
        return self.task.app

    @property
    def __iter__(self) -> Iterator[T]:
        yield from self.invocations

    @property
    @abstractmethod
    def results(self) -> Iterator[Result]:
        """"""
